


"""
This module defines the page objects
"""


import os, mimetypes, copy, collections, json, re, uuid, pprint
from string import Template
from urllib.parse import quote
from base64 import urlsafe_b64encode

from . import skiboot
from .tag import Part, ClosedPart, TextBlock, Section
from .widgets import links
from .excepts import ValidateError, ServerError, SkiError

# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')


class ParentPage(object):

    def __init__(self, name="", brief = "New Page"):
        """Initiates a ParentPage instance

        name: a url friendly page name
        """
        # self._parentfolder_ident is set when a page is added to a folder.
        self._parentfolder_ident = None
        if name:
            self._name = name.lower()
        else:
            self._name = ""
        self.ident = None
        self.status = '200 OK'
        self.headers = [('content-type', 'text/html')]
        self.brief = brief
        # the change is a uuid which alters whenever the page changes
        self.change = uuid.uuid4().hex
        self.ident_data = ''
        # Set by end_call
        self.session_cookie = ()
        self.language_cookie = ()

    def get_name(self):
        "The page name"
        return self._name

    def set_name(self, name):
        "Ensure name is lower case"
        self._name = name.lower()

    name = property(get_name, set_name)

    @property
    def page_type(self):
        return self.__class__.__name__

    @property
    def ident_data_string(self):
        "Encode ident_data as a base64 string"
        if not self.ident:
            return
        if self.ident_data:
            # encode the ident data as base64, and append it as a string to the page ident
            b64binarydata = urlsafe_b64encode(self.ident_data.encode('ascii')).rstrip(b"=")  # removes final '=' padding
            return str(self.ident) + '_' + b64binarydata.decode('ascii')
        else:
            return str(self.ident)        

    @property
    def proj_ident(self):
        return self.ident.proj

    @property
    def project(self):
        if self.ident:
            return skiboot.getproject(proj_ident=self.ident.proj) 

    def copy(self, name, brief):
        "Returns a deepcopy of this page, with new name and brief, parentfolder and ident are None"
        copy_page = copy.deepcopy(self.project.get_item(self.ident))
        copy_page._parentfolder_ident = None
        copy_page.ident = None
        copy_page.name = name
        copy_page.brief = brief
        copy_page.change = uuid.uuid4().hex
        return copy_page

    def root(self):
        "Returns the page root folder"
        if self.ident():
            proj = self.project()
            if proj is None:
                return
            return proj.root


    def get_parentfolder(self):
        "Uncopied parent folder"
        if self._parentfolder_ident is None:
            return None
        return skiboot.get_item(self._parentfolder_ident)

    def set_parentfolder(self, parentfolder):
        "Warning, this does not set this page into the parents pages dictionary, it just sets this parentfolder attribute"
        if parentfolder is None:
            self._parentfolder_ident = None
        else:
            self._parentfolder_ident = parentfolder.ident

    parentfolder = property(get_parentfolder, set_parentfolder)

    def get_folder(self):
        "folder is an alias of parentfolder"
        return self.get_parentfolder()

    def set_folder(self, parentfolder):
        "folder is an alias of parentfolder"
        self.set_parentfolder(parentfolder)

    folder = property(get_folder, set_folder)

    @property
    def parentfolder_ident(self):
        "Returns the parentfolder ident"
        return self._parentfolder_ident

    @property
    def url(self):
        if self._parentfolder_ident is None:
            return ""
        return self._parentfolder_ident.url() + self.name


    def parent_list(self):
        "returns a list of (name, identnumber) starting at root"
        p_list = self.parentfolder.parent_list()
        p_list.append((self.name, self.ident.num))
        return p_list


    @property
    def restricted(self):
        "returns restricted status of parentfolder"
        parentfolder = self.parentfolder
        if parentfolder is None:
            return True
        return parentfolder.restricted

    def update(self, environ, call_data, lang, ident_list=[]):
        "override"
        if self.session_cookie:
            self.headers.append(self.session_cookie)
        if self.language_cookie:
            self.headers.append(self.language_cookie)

    def show_error(self, error_messages=None):
        """override"""
        pass

    def set_values(self, page_data):
        """Checks for special page template values"""
        if ('status' in page_data):
            if page_data['status']:
                self.status = page_data['status']
            del page_data['status']
        if ('headers' in page_data):
            if page_data['headers']:
                self.headers = page_data['headers']
            del page_data['headers']
        if ('ident_data' in page_data):
            if page_data['ident_data']:
                ident_data = page_data['ident_data']
                if not isinstance(ident_data, str):
                    ident_data = str(ident_data)
                # set ident_data in page
                self.ident_data = ident_data
            del page_data['ident_data']

    def import_sections(self, page_data=None):
        "Only used by Template and SVG, everything else just returns"
        return


    def get_status(self):
        "Returns (status, headers)"
        return self.status, self.headers

    def __eq__(self, other):
        return self.ident == other.ident

    def __ne__(self, other):
        return self.ident != other.ident

    def __bool__(self):
        return True


class TemplatePageAndSVG(ParentPage):
    "Acts as parent to template pages and svg pages"


    def __init__(self, name="", brief = "New Page"):
        """Initiates a TemplatePageAndSVG instance
        """
        ParentPage.__init__(self, name=name, brief=brief)
        # self.widgets is a widget name -> widget dictionary
        self.widgets = {}
        # self.section_places is a page section name -> SectionPlaceHolder dictionary
        self.section_places = {}
        # dictionary of sections, page section name -> section
        # this is filled in when the sections are imported
        self.sections = {}


    def set_values(self, page_data):
        """Sets the widget fields to the given values.
           page_data is either a dictionary of
           {(pagesectionname,widgetname,fieldname):value, ..}
           or a dictionary of
           {'pagesectionname-widgetname:fieldname':value, ..}
           or a dictionary of
          {WidgField_object:value, ..}
           or a dictionary of
          {pagesectionname:value, ..}

           NOTE : currently untested, but this should work with index as well
           so if the widgfield is of format (pagesectionname,widgetname,fieldname,index)
           or 'pagesectionname-widgetname:fieldname-index' then the single indexed field value
           should be updated
           """
        ParentPage.set_values(self, page_data)
        for widgfield, value in page_data.items():
            # could be the special case of setting a section show or class value
            if self._check_section_parameters(widgfield, value):
                continue
            widget, fieldname = self.widget_from_field(widgfield)
            if (widget is not None) and fieldname:
                widget.set_value(fieldname, value)


    def _check_section_parameters(self, widgfield, value):
        """Checks special case of ('pagesectionname','show'),
                               or ('pagesectionname','hide')
                               or ('pagesectionname','section_class')"""
        if isinstance(widgfield, str) and (':' in widgfield):
            widglist = widgfield.split(':')
            if len(widglist) != 2:
                return False
            sectionname, field = widglist
        elif (isinstance(widgfield, list) or isinstance(widgfield, tuple)) and (len(widgfield) == 2):
            sectionname, field = widgfield
        else:
            return False
        if sectionname not in self.sections:
            return False
        if (field == 'show') and ((value is True) or (value is False)):
            self.sections[sectionname].show = value
            return True
        if field == 'section_class':
            self.sections[sectionname].set_class(value)
            return True
        if field == 'hide':
            self.sections[sectionname].set_hide(value)
            return True
        return False


    def widget_from_field(self, widgfield):
        """Return the (widget, fieldname) given a widgfield, which is either a string of
         pagesectionname-widgetname:fieldname or a list/tuple or cfg.WidgField object, if widget not found in the page,
         (None, fieldname) will be returned.
        If widget is found, but the fieldname is not in the widget, return (widget, None)
        The widget returned is the actual widget instance in the page, and fieldname is the widgets field name
        Please note - if the given widgfield has an index, fieldname will be 'fieldname-index'"""
        if not widgfield:
            raise ValidateError(message = "Invalid field given to page.widget_from_field method")
        widg_field = skiboot.make_widgfield(widgfield)
        # widg_field is now a WidgField object with s, w, f and i attributes
        if (not widg_field.w) or (not widg_field.f):
            raise ValidateError(message = "Field %s not recognised" % (widgfield,))
        widget = self.widget_from_name(widg_field.s, widg_field.w)
        if widg_field.f == 'show_error':
            return widget, 'show_error'
        # check is the fieldname in the widget
        if widget is not None:
            if not widget.is_fieldname_in_widget(widg_field.f):
                return widget, None
        if widg_field.i:
            fieldname = widg_field.f + '-' + widg_field.i
        else:
            fieldname = widg_field.f
        return widget, fieldname

    def widget_from_name(self, section_name, widgetname):
        """Return the widget given a widgetname, if widget not found in the page, None will be returned
           This should only be called after sections have been imported and returns the actual widget in the page"""
        if not section_name:
            # widget is not in a section, should be local to this page
            return self.widgets.get(widgetname)
        if section_name not in self.sections:
            return None
        sectionpart = self.sections[section_name]
        return sectionpart.widgets.get(widgetname)


    def copy_widget_from_name(self, section_name, widgetname):
        """Return the widget given a widgetname, if widget not found in the page or section, None will be returned
           This can be called before or after sections have been imported and returns a copy of the widget"""
        if not section_name:
            # widget is not in a section, should be local to this page
            widget = self.widgets.get(widgetname)
        elif section_name in self.sections:
            section = self.sections[section_name]
            widget = section.widgets.get(widgetname)
        elif section_name in self.section_places:
            # self.section_places is a page section name -> SectionPlaceHolder dictionary
            sectionplaceholder = self.section_places.get(section_name)
            if not sectionplaceholder:
                return
            section = sectionplaceholder.get_section()
            widget = section.widgets.get(widgetname)
        elif "_" in section_name:
            result = self.widget_from_multiplier(section_name, widgetname)
            if result:
                sectionplaceholder,widget = result
            else:
                widget = None
        else:
            return
        if widget is None:
            return
        return copy.deepcopy(widget)


    def widget_from_multiplier(self, section_name, widgetname):
        """Return (sectionplaceholder, widget) if section_name is a multiplied section, such as name_3, return None on failure"""
        if "_" not in section_name:
            return
        sname, snumber = section_name.rsplit('_', 1)
        try:
            snumber = int(snumber)
        except Exception:
            return
        sectionplaceholder = self.section_places.get(sname)
        if not sectionplaceholder:
            return
        section = sectionplaceholder.get_section()
        return sectionplaceholder, section.widgets.get(widgetname)


    def append_scriptlink(self, label):
        """Used to append a scriptlink to head, overridden in class Page"""
        pass

    def import_sections(self, page_data=None):
        "Imports the project sections into the page"
        self.sections = {}
        for placename, placeholder in self.section_places.items():
            # get top part of placeholder location
            pagepart = placeholder.pagepart
            if pagepart == 'head':
                toppart = self.head
            elif pagepart == 'body':
                toppart = self.body
            elif pagepart == 'svg':
                toppart = self.svg
            else:
                raise ServerError(message = 'Invalid section placeholder in page %s' % (self.ident,))
            if page_data and (placename,'multiplier') in page_data:
                try:
                    multiplier = int(page_data[placename,'multiplier'])
                except Exception:
                    placeholder.multiplier = 0
                else:
                    placeholder.multiplier = multiplier
            if placeholder.multiplier > 0:
                for m in range(placeholder.multiplier):
                    self._import_multiplied_section(m, placeholder, toppart, page_data)
                continue
            section_name = placeholder.section_name
            sectionpart = self.project.section(section_name)
            # sectionpart is a tag.Section
            if isinstance(sectionpart, Section):
                # gives sectionpart and subparts an ident of page_ident_name_locationnumbers
                # and sets sectionpart into self.sections
                sectionpart.widgets = {}
                sectionpart.section_places = {}
                sectionpart.set_idents(str(self.ident) + '_' + placename, sectionpart.widgets, sectionpart.section_places, embedded=(section_name,'',None))
                # If no id placed in the top tag, inserts the section placename
                if not sectionpart.has_attrib('id'):
                    sectionpart.insert_id(id_string=placename)
                self.sections[placename] = sectionpart
            else:
                continue
            # Set section widgets field displaynames
            for widget in sectionpart.widgets.values():
                # Note - this is called on named widgets only
                widget.set_placename(section_name, placename)
            # insert sectionpart at the point where the placeholder is
            toppart.set_location_value(placeholder.ident_list, sectionpart)
            # add section scriptlinks
            if self.page_type == 'TemplatePage':
                if sectionpart.validator_scriptlinks:
                    for link_label in sectionpart.validator_scriptlinks:
                        self.append_scriptlink(link_label)
                for widget in sectionpart.widgets.values():
                    # get 'ski_modulename'
                    link_label = "ski_" + widget.__class__.__module__.split(".")[-1]
                    # add a link in the page head to the widget module javascript file
                    self.append_scriptlink(link_label)


    def _import_multiplied_section(self, m, placeholder, toppart, page_data):
        "If a placeholder has a multiplier, import its section multiple times inside a div"
        placename = placeholder.placename + "_" + str(m)
        section_name = placeholder.section_name
        sectionpart = self.project.section(section_name)
        # sectionpart is a tag.Section
        if isinstance(sectionpart, Section):
            # gives sectionpart and subparts an ident of page_ident_name_locationnumbers
            # and sets sectionpart into self.sections
            sectionpart.widgets = {}
            sectionpart.section_places = {}
            sectionpart.set_idents(str(self.ident) + '_' + placename, sectionpart.widgets, sectionpart.section_places, embedded=(section_name,'',None))
            # If no id placed in the top tag, inserts the section placename
            if not sectionpart.has_attrib('id'):
                sectionpart.insert_id(id_string=placename)
            self.sections[placename] = sectionpart
        else:
            return
        # Set section widgets field displaynames
        for widget in sectionpart.widgets.values():
            # Note - this is called on named widgets only
            widget.set_placename(section_name, placename)
        # now the sectionpart has to be set within a div which is set at the placeholder location
        if m == 0:
            topdiv = Part(tag_name=placeholder.mtag)
            if page_data:
                if (placeholder.placename,'show') in page_data:
                    topdiv.show = bool(page_data[placeholder.placename,'show'])
                if (placeholder.placename,'section_class') in page_data:
                    topdiv.set_class(page_data[placeholder.placename,'section_class'])
                if (placeholder.placename,'multiplier_tag') in page_data:
                    topdiv.tag_name = page_data[placeholder.placename,'multiplier_tag']
                if self.page_type == 'TemplatePage':
                    # hide not relevant to svg
                    if (placeholder.placename,'hide') in page_data:
                        topdiv.set_hide(bool(page_data[placeholder.placename,'hide']))
            topdiv.insert_id(id_string=placeholder.placename)
            topdiv[0] = sectionpart
            toppart.set_location_value(placeholder.ident_list, topdiv)
        else:
            location = placeholder.ident_list[:]
            location.append(m)
            toppart.insert_location_value(location, sectionpart)
        # add section scriptlinks
        if self.page_type == 'TemplatePage':
            if sectionpart.validator_scriptlinks:
                for link_label in sectionpart.validator_scriptlinks:
                    self.append_scriptlink(link_label)
            for widget in sectionpart.widgets.values():
                # get 'ski_modulename'
                link_label = "ski_" + widget.__class__.__module__.split(".")[-1]
                # add a link in the page head to the widget module javascript file
                self.append_scriptlink(link_label)


class TemplatePage(TemplatePageAndSVG):
    """A template page object
    """

    def __init__(self, name="",
                       brief = "New Page",
                       show_backcol=False,
                       backcol="#FFFFFF",
                       last_scroll=True,
                       default_error_widget=None,
                       lang=None,
                       interval = 0,
                       interval_target=None,
                       catch_to_html=None):
        """Initiates a Page instance

        name: a url friendly page name
        brief: description of the page
        show_backcol: True if colour can be placed in the html tag
        backcol: The colour to place in the html tag
        last_scroll: True if the page is to remember its last scroll position
        default_error_widget: widget name to display errors, if no other widget specified in the error call
        lang: language string to place in the html tag
        interval: Time interval in integer seconds for repeated call for JSON update, 0 if not used.
        interval_target: ident or label (not URL) to call for the JSON update
        catch_to_html: ident or label (not URL) to call when a javascript error is thrown
        """

        TemplatePageAndSVG.__init__(self, name=name, brief=brief)
        # set up page head
        self.head = Part(tag_name="head")
        self.head.brief = "The head section of the page"
        # Set up the page body
        self.body = Part(tag_name="body")
        self.body.brief = "The body section of the page"

        # html lang setting
        if lang is None:
            self.lang = skiboot.default_language()
        else:
            self.lang = lang

        # ERROR MESSAGES
        # This widget is the default widget where errors are displayed, if no other error location is specified
        if default_error_widget is None:
            self._default_error_widget = skiboot.WidgField(s='', w="top_error", f='', i='')
        else:
            self.default_error_widget = default_error_widget
        # flag if any error present
        self.page_in_error = False
        # store errors as they arrive
        # self.stored_errors will be a dictionary with key widgetname (of the widget to show the error)
        # and value being the message to be displayed by the widget
        self.stored_errors = {}

        self.show_backcol = show_backcol
        self.backcol = backcol
        self.catch_to_html = catch_to_html

        # A list of appended script links in head
        self._headlinks = []

        # Used to hold javascript created at run time
        self._js = ''

        # A javscript string for each widgfield
        self._scriptcontents = ''

        # last scroll is True if the page is to be restored to the last scrolled position
        self.last_scroll = last_scroll

        # if requesting a json target to be called at regular intervals,
        # self.interval is the time in seconds, and self.interval_target
        # is the target url
         
        self.interval = interval
        if interval_target:
            self.interval_target = interval_target
        else:
            self.interval_target = None

        # self.validator_scriptlinks is a list of validator module ski_name's
        # calculated when the page is saved, and used 
        # to add links to the validator js modules in the page head
        self._validator_scriptlinks = []

        # These will be added to the page javascript within the jquery $(document).ready(function() {
        self._add_storage = ''
        self._add_jscript = ''

    @property
    def validator_scriptlinks(self):
        return self._validator_scriptlinks

    def load_validator_scriptlinks(self):
        "Called when page saved to database"
        self._validator_scriptlinks = []
        val_modules = []
        for widget in self.widgets.values():
            for field in widget.fields.values():
                val_list = field.val_list
                for val in val_list:
                    mod_name = val.__class__.__module__.split(".")[-1]
                    if mod_name not in val_modules:
                        val_modules.append(mod_name)
        for mod_name in val_modules:
            link_label = 'ski_' + mod_name
            if link_label not in self._validator_scriptlinks:
                self._validator_scriptlinks.append(link_label)

    def import_sections(self, page_data=None):
        "Imports javascript modules used by widgets and validators, and then imports sections"
        if self._validator_scriptlinks:
            for link_label in self._validator_scriptlinks:
                        self.append_scriptlink(link_label)
        for widget in self.widgets.values():
            # get 'ski_modulename'
            link_label = "ski_" + widget.__class__.__module__.split(".")[-1]
            # add a link in the page head to the widget module javascript file
            self.append_scriptlink(link_label)
        # import sections using parent method
        TemplatePageAndSVG.import_sections(self, page_data)



    def location_item(self, location):
        "Returns the part or widget at location"
        location_string, container_number, location_list = location
        if not location_string:
            return
        if container_number is None:
            # Not in a widget
            if location_string == 'body':
                if not location_list:
                    return self.body
                return self.body.get_location_value(location_list)
            elif location_string == 'head':
                if not location_list:
                    return self.head
                return self.head.get_location_value(location_list)
            # not found
            return
        # location string must be a widget name
        # so part is within a widget container within the page
        if location_string not in self.widgets:
            return
        widget = self.widgets.get(location_string)
        if widget is None:
            return
        if widget.can_contain():
            if not location_list:
                # return the container
                return widget.container_part(container_number)
            else:
                return widget.get_from_container(container_number, location_list)
        # part not found
        return

    def set_default_error_widget(self, errorwidget):
        "Widget only, no field information"
        self._default_error_widget = skiboot.make_widgfield(errorwidget, widgetonly=True)

    def get_default_error_widget(self):
        return self._default_error_widget

    default_error_widget = property(get_default_error_widget, set_default_error_widget)

    def append_scriptlink(self, label):
        """Used to append a scriptlink to head"""
        if label not in self._headlinks:
            self._headlinks.append(label)
            # create a scriptlink
            scriptlink = Part(tag_name = "script", attribs={"src":"{" + label + "}"}, brief='script link to %s' % (label,))
            self.head.append(scriptlink)

    def add_javascript(self, contents):
        "Adds contents to self._scriptcontents"
        self._scriptcontents += contents


    def make_js(self, ident_list, environ, call_data, lang):
        "Auto generates the javascript for this page, sets into self._js, called after head and body updates"
        # for each part in the page body, get its associated javascript, and add to scriptcontents
        # send self, plus call arguments so part.make_js has all data necessary
        self.body.make_js(self, ident_list, environ, call_data, lang)
        # The start of the script is created in the .data function
        scriptmiddle = """
// Widget functions
$(document).ready(function(){
"""
        if self._add_storage:
            scriptmiddle += self._add_storage

        # and create the script end
        scriptend = self._add_jscript + """
  if(SKIPOLE.interval && SKIPOLE.IntervalTarget) {
    SKIPOLE.interval_id = setInterval(SKIPOLE.refreshjson, SKIPOLE.interval*1000, SKIPOLE.IntervalTarget);
    }
"""
        # SKIPOLE.interval_id is needed so clearInterval(SKIPOLE.interval_id) can be called
        # if the interval changes
        if self.last_scroll:
            # restore to last store position
            scriptend += """
  SKIPOLE.restorepagepos();
"""
        # and set an event to store window position on unload
        scriptend += """
  if(typeof(Storage) !== "undefined") {
    $(window).on('unload', function(){
      // on unloading, store the x and y position of the window
      sessionStorage.setItem('%s_y', $(window).scrollTop().toString());
      sessionStorage.setItem('%s_x', $(window).scrollLeft().toString());
      });
   }
});
""" % (self.ident, self.ident)
        self._js = scriptmiddle + self._scriptcontents + scriptend


    def update(self, environ, call_data, lang, ident_list=[]):
        "Updates page and body parts, lang is a tuple of (language, default_language)"
        # show all errors on the page
        self._show_errors_on_update()
        self.body.update(self, ident_list, environ, call_data, lang, str(self.ident)+"_body")
        # update page head parts (done last as previous updates may change css)
        self.head.update(self, ident_list, environ, call_data, lang, str(self.ident)+"_head")
        # create javascript for widgets, this is set into self._js and appended
        # to self.head in the data function
        self.make_js(ident_list, environ, call_data, lang)
        if self.session_cookie:
            self.headers.append(self.session_cookie)
        if self.language_cookie:
            self.headers.append(self.language_cookie)


    def set_idents(self):
        """Each widget adds its ident, to itself, and to the self.widgets dictionary
           Each sectionplaceholder adds its ident, to itself, and to the self.section_places dictionary
           This is called when the page is created and added to a folder, it is also
           called when an existing page is edited, and the page saved"""
        self.widgets = {}
        self.section_places = {}
        self.body.set_idents(str(self.ident)+"_body", self.widgets, self.section_places)
        self.head.set_idents(str(self.ident)+"_head", self.widgets, self.section_places)


    def validate(self, widgfield, value, environ, lang, form_data, call_data, page_data):
        """Given a widgfield, its value and other parameters, call its widget validate function
           which returns (value, error_list)"""
        if not widgfield:
            raise ValidateError(message = "Invalid widget-field, unable to validate")
        widg_field = skiboot.make_widgfield(widgfield)

        # widg_field is now a WidgField object with s, w, f and i attributes
        if (not widg_field.w) or (not widg_field.f):
            raise ValidateError(message = "Field %s not recognised" % (widgfield,))

        # Find the widget in this page, to call its validate function
        widget = None

        if widg_field.s:
            # self.section_places is a page section name -> SectionPlaceHolder dictionary
            sectionplaceholder = self.section_places.get(widg_field.s)
            if sectionplaceholder:
                sectionpart = sectionplaceholder.get_section()
                widget = sectionpart.widgets.get(widg_field.w)
            elif "_" in widg_field.s:
                result = self.widget_from_multiplier(widg_field.s, widg_field.w)
                if result:
                    sectionplaceholder, widget = result                
        else:
            # widget is not in a section, should be local to this page
            widget = self.widgets.get(widg_field.w)

        if widget is None:
            raise ValidateError(message = "Widget %s not found in page %s" % (widg_field.w, self.ident))

        # check is the fieldname in the widget
        if not widget.is_fieldname_in_widget(widg_field.f):
            raise ValidateError(message = "Field %s not found in widget %s, page %s" % (widg_field.f, widg_field.w, self.ident))

        # validate
        newvalue, error_list = widget.validate(widg_field, value, environ, lang, form_data, call_data, self.ident)

        # errors in same section as the errored widget need s value changing
        if widg_field.s:
            for error in error_list:
                if error.section == sectionplaceholder.section_name:
                    error.section = widg_field.s

        return newvalue, error_list


    def show_error(self, error_messages=[]):
        """Stores the errors to display on the page in dictionary self.stored_errors.
           These are subsequently set on the widgets during the page update by _show_errors_on_update().
           Any fields which are just to be flagged as errored, via the widgets mark_field_in_error are set here.
           error_messages is a list of ErrorMessage instances (class defined in excepts)"""
        # an ErrorMessage has attributes message,
        #                                section,
        #                                widget,
        self.page_in_error = True
        for e in error_messages:
            # e is an instance of ErrorMessage
            if e.widget:
                displaywidget = skiboot.WidgField(s=e.section, w=e.widget, f='', i='')
            else:
                # use page default error location 
                displaywidget = self.default_error_widget
            if not displaywidget:
                # nowhere to display an error
                continue
            if displaywidget not in self.stored_errors:
                self.stored_errors[displaywidget] = e.message
            else:
                # displaywidget is already in self.stored_errors
                # existing messages have priority
                # unless existing message is an empty string
                if not self.stored_errors[displaywidget]:
                    self.stored_errors[displaywidget] = e.message

                
    def _show_errors_on_update(self):
        """Called on page update to show all stored errors"""
        if not self.page_in_error:
            # show_error never called - so no error condition
            return
        # if no error widget or message given, show the page default error widget
        if not self.stored_errors:
            widget = self.widget_from_name(self.default_error_widget.s, self.default_error_widget.w)
            if widget is None:
                return
            widget.show_error(message='')
            return
        # There are stored messages, so call the errored widgets
        for displaywidget, message in self.stored_errors.items():
            if not displaywidget:
                continue
            widget = self.widget_from_name(displaywidget.s, displaywidget.w)
            if widget is None:
                continue
            widget.show_error(message)

    def set_values(self, page_data):
        """Checks for special page template values, then passes on to parent set_values to set the widgets"""
        try:
            if 'set_cookie' in page_data:
                # sets the cookies in the page headers
                sendcookies = page_data['set_cookie']
                if sendcookies:
                    for morsel in sendcookies.values():
                        self.headers.append(("Set-Cookie", morsel.OutputString()))
                del page_data['set_cookie']                      # delete to stop parent object from testing this item is a widgfield
            if 'CatchToHTML' in page_data:
                self.catch_to_html = page_data['CatchToHTML']
                del page_data['CatchToHTML']                      # delete to stop parent object from testing this item is a widgfield
            if 'interval' in page_data:
                interval = 0
                try:
                    interval = page_data['interval']
                    del page_data['interval']
                    interval = int(interval)
                except:
                    pass
                else:
                    self.interval=interval
            if 'IntervalTarget' in page_data:
                self.interval_target = page_data['IntervalTarget']
                del page_data['IntervalTarget']
            if 'last_scroll' in page_data:
                self.last_scroll = bool(page_data['last_scroll'])
                del page_data['last_scroll']
            if 'lang' in page_data:
                if isinstance(page_data['lang'], 'tuple') or isinstance(page_data['lang'], 'list'):
                    self.lang = page_data['lang'][0]
                else:
                    self.lang = page_data['lang']
                del page_data['lang']
            if 'show_backcol' in page_data:
                self.show_backcol = bool(page_data['show_backcol'])
                del page_data['show_backcol']
            if 'backcol' in page_data:
                self.backcol = page_data['backcol']
                del page_data['backcol']
            if 'body_class' in page_data:
                self.body.update_attribs({'class':page_data['body_class']})
                del page_data['body_class']
            if 'show_error' in page_data:
                page_data[self._default_error_widget.set_field(f='show_error')] = page_data['show_error']
                del page_data['show_error']
            if 'add_jscript' in page_data:
                self._add_jscript = page_data['add_jscript']
                del page_data['add_jscript']
            if ('localStorage' in page_data) or ('sessionStorage' in page_data):
                self._add_storage += """  if (typeof(Storage) !== "undefined") {
"""
                if 'localStorage' in page_data:
                    if not isinstance(page_data['localStorage'], dict):
                        raise ServerError("localStorage must be a dictionary")
                    for key,val in page_data['localStorage'].items():
                        escapedval = json.dumps(val)
                        self._add_storage += """    localStorage.setItem("%s", %s);
""" % (key,escapedval)
                    del page_data['localStorage']
                if 'sessionStorage' in page_data:
                    if not isinstance(page_data['sessionStorage'], dict):
                        raise ServerError("sessionStorage must be a dictionary")
                    for key,val in page_data['sessionStorage'].items():
                        escapedval = json.dumps(val)
                        self._add_storage += """    sessionStorage.setItem("%s", %s);
""" % (key,escapedval)
                    del page_data['sessionStorage']
                self._add_storage += """    }
"""
        except ServerError:
            raise
        except:
            raise ServerError("Error while setting values into Template page")
        TemplatePageAndSVG.set_values(self, page_data)

    def data(self):
        "Returns the page as a list of binary strings"
        # set data into the page
        page_javascript = """// javascript for skipole project %s page ident %s
//<![CDATA[
SKIPOLE.identdata = '%s';
SKIPOLE.default_error_widget = '%s';
SKIPOLE.widget_register = {};
SKIPOLE.interval = %s;
"""  % (self.ident.proj, self.ident.num, self.ident_data_string, self.default_error_widget, self.interval)
        if self.interval_target:
            intervalurl = skiboot.get_url(self.interval_target, self.proj_ident)
            if intervalurl:
                page_javascript += "SKIPOLE.IntervalTarget = \'" + quote(intervalurl, safe='/:?&=') + "\';\n"
        if self.catch_to_html:
            catchurl = skiboot.get_url(self.catch_to_html, self.proj_ident)
            if catchurl:
                page_javascript += "SKIPOLE.CatchToHTML = \'" + quote(catchurl, safe='/:?&=') + "\';\n"

        # set a list of section alias names, to include section alias, and multiplied section alias
        section_alias_list = list(self.sections.keys())
        for alias in self.section_places.keys():
            if alias not in section_alias_list:
                section_alias_list.append(alias)

        page_javascript += "SKIPOLE.sections =  " + str(section_alias_list) + ";\n"

        # append javascript to page head
        self.head.append(Part(tag_name="script", text=page_javascript + self._js + "//]]>\n"))

        topstring = "<!DOCTYPE HTML>\n<html"
        if self.show_backcol:
            back = " style=\"background-color: %s;\"" % self.backcol
            topstring += back
        if self.lang:
            topstring += " lang=\"%s\"" % self.lang
        topstring += ">\n"
        topbytes = topstring.encode('ascii', 'xmlcharrefreplace')

        docbytes = [topbytes]
        docbytes.extend(d.encode('ascii', 'xmlcharrefreplace') for d in self.head.data())
        docbytes.extend(d.encode('ascii', 'xmlcharrefreplace') for d in self.body.data())
        docbytes.append("\n</html>".encode('ascii', 'xmlcharrefreplace'))
        return docbytes

    def __str__(self):
        "Returns the page as a string"
        topstring = "<!DOCTYPE HTML>\n<html"
        if self.show_backcol:
            back = " style=\"background-color: %s;\"" % self.backcol
            topstring += back
        if self.lang:
            topstring += " lang=\"%s\"" % self.lang
        topstring += ">"
        return topstring + str(self.head) + str(self.body) + "\n</html>"

    def __repr__(self):
        if self.ident:
            return "Template Page ident %s" % (self.ident,)
        else:
            return "Template Page name %s" % (self.name,)


class RespondPage(ParentPage):


    def __init__(self, name, brief = "New RespondPage", responder=None):
        ParentPage.__init__(self, name=name, brief=brief)
        # A responder is an instance of the Respond class
        # - or more usually, an instance of a child of the Respond class
        self.responder = responder

    def call_responder(self, skicall, form_data, caller_ident, ident_list, rawformdata):
        """Checks for circulating calls, then updates ident_list with this pages ident,
           then calls the respond objects call method, note rawformdata is a cgi.FieldStorage object"""
        if self.responder is None:
            raise ServerError(message = "No responder has been assigned to this page")
        if self.ident in ident_list:
            raise ServerError(message="Circulating call to Respond page: ident %s" % (self.ident,))
        ident_list.append(self.ident)
        proj = self.ident.proj

        # create submit_dict
        skicall.submit_dict["responder_brief"] = self.brief
        skicall.submit_dict["number"] = self.ident.num

        page = self.responder(skicall, form_data, caller_ident, ident_list, proj, rawformdata)
        return page

    def set_values(self, page_data):
        return

    def __repr__(self):
        if self.ident:
            return "Respond Page ident %s" % (self.ident,)
        else:
            return "Respond Page name %s" % (self.name,)



class SVG(TemplatePageAndSVG):
    """An SVG image
    """

    def __init__(self, name="", brief = "New SVG Page", width="100", height="100", css_list=[], enable_cache=False):
        """Initiates a Page instance

        name: a url friendly page name
        """
        TemplatePageAndSVG.__init__(self, name=name, brief=brief)
        self.css_list = [skiboot.make_ident(ident) for ident in css_list]
        self.svg = Part(tag_name="svg", attribs = {"xmlns":"http://www.w3.org/2000/svg",
                                                   "baseProfile":"full",
                                                   "version":"1.1",
                                                   "width":width,
                                                   "height":height})
        self.svg.brief = "The svg section of the page"
        # enable_cache flag
        self._enable_cache = False
        self.enable_cache = enable_cache

    def set_enable_cache(self, enable_cache):
        "Sets enable cache in header"
        if enable_cache:
            self._enable_cache = True
            self.headers = [('content-type', 'image/svg+xml'), ('cache-control', 'max-age=3600')]
        else:
            self._enable_cache = False
            self.headers = [('content-type', 'image/svg+xml'),
                                              ('cache-control','no-cache, no-store, must-revalidate'),
                                              ('Pragma', 'no-cache'),
                                              ( 'Expires', '0') ]


    def get_enable_cache(self):
        return self._enable_cache

    enable_cache = property(get_enable_cache, set_enable_cache)


    def location_item(self, location):
        "Returns the part or widget at location"
        location_string, container_number, location_list = location
        if not location_string:
            return
        if container_number is None:
            # Not in a widget
            if location_string == 'svg':
                if not location_list:
                    return self.svg
                return self.svg.get_location_value(location_list)
            # not found
            return
        # location string must be a widget name
        # so part is within a widget container within the page
        if location_string not in self.widgets:
            return
        widget = self.widgets.get(location_string)
        if widget is None:
            return
        if widget.can_contain():
            if not location_list:
                # return the container
                return widget.container_part(container_number)
            else:
                return widget.get_from_container(container_number, location_list)
        # part not found
        return



    def get_width(self):
        return self.svg.get_attrib_value("width")

    def set_width(self, width):
        self.svg.update_attribs({"width":width})

    width = property(get_width, set_width)

    def get_height(self):
        return self.svg.get_attrib_value("height")

    def set_height(self, height):
        self.svg.update_attribs({"height":height})

    height = property(get_height, set_height)

    def update(self, environ, call_data, lang, ident_list=[]):
        "Updates svg parts"
        ident_string = str(self.ident)+"_svg"
        self.svg.update(self, ident_list, environ, call_data, lang, ident_string)
        if self.session_cookie:
            self.headers.append(self.session_cookie)
        if self.language_cookie:
            self.headers.append(self.language_cookie)

    def add_javascript(self, contents):
        "Currently not used"
        pass

    def set_idents(self):
        """Each widget adds its ident, to itself, and to the self.widgets dictionary
           Each sectionplaceholder adds its ident, to itself, and to the self.section_places dictionary
           This is called when the page is created and added to a folder, it is also
           called when an existing page is edited"""
        self.widgets = {}
        self.section_places = {}
        self.svg.set_idents(str(self.ident)+"_svg", self.widgets, self.section_places)

    def show_error(self, error_messages=None):
        """Currently not used"""
        return


    def set_values(self, page_data):
        """Checks for special svg page template values, then passes on to parent set_values to set the widgets"""
        if 'enable_cache' in page_data:
            self.enable_cache = bool(page_data['enable_cache'])
            del page_data['enable_cache']
        if 'width' in page_data:
            self.width = page_data['width']
            del page_data['width']
        if 'height' in page_data:
            self.height = page_data['height']
            del page_data['height']
        TemplatePageAndSVG.set_values(self, page_data)


    def data(self):
        "Returns the page as a list of binary strings"

        # note xmls initial dtd statements and stuff left out as per
        # https://developer.mozilla.org/en-US/docs/Web/SVG/Tutorial/Getting_Started

        svg_list = self.svg.data()
        return [d.encode('ascii', 'xmlcharrefreplace') for d in svg_list]


    def __str__(self):
        return str(self.svg)

    def __repr__(self):
        if self.ident:
            return "SVG Page ident %s" % (self.ident,)
        else:
            return "SVG Page name %s" % (self.name,)


class FilePage(ParentPage):
    """A page representing a file to be downloaded to the user.
    """

    def __init__(self, name="", brief="", filepath="", mimetype=None, enable_cache=False):
        "If no filepath given, assumes a file of the same name under the project static folder"
        if not brief:
            brief = "Link to %s" % filepath
        ParentPage.__init__(self, name=name, brief=brief)
        self.filepath = filepath
        self._filepath_relative_to_project_files = None
        self._mimetype = mimetype
        self._enable_cache = False
        self.enable_cache = enable_cache
        # flag for headers auto set
        self._headers_flag = True
        # environ, set by the update methon
        self._environ = None

    def set_enable_cache(self, enable_cache):
        "Sets enable cache in header"
        if enable_cache:
            self._enable_cache = True
            self.headers = [('cache-control', 'max-age=3600')]
        else:
            self._enable_cache = False
            self.headers = [ ('cache-control','no-cache, no-store, must-revalidate'),
                                              ('Pragma', 'no-cache'),
                                              ( 'Expires', '0')]

    def get_enable_cache(self):
        return self._enable_cache

    enable_cache = property(get_enable_cache, set_enable_cache)

    def get_filepath(self):
        # self._filepath is a list of path parts
        if (not self._filepath):
            self._filepath = [self.proj_ident, "static"]
            self._filepath.append(self.name)
        return os.sep.join(self._filepath)

    def set_filepath(self, path):
        if path:
            if ('/' in path) and ('\\' in path):
                raise ValidateError("Invalid path")
            if '/' in path:
                self._filepath = path.strip('/').split('/')
            elif '\\' in path:
                self._filepath = path.strip('\\').split('\\')
            else:
                self._filepath = [path]
        else:
            self._filepath = []
        for item in self._filepath:
            if (not item) or (item == '.') or (item == '..'):
                self._filepath = []
                raise ValidateError("Invalid path")

    filepath = property(get_filepath, set_filepath)

    def get_mimetype(self):
        if self._mimetype: return self._mimetype
        # self._mimetype not set, so guess it
        if self.filepath:
            name = os.path.basename(self.filepath)
        else:
            name=self.name
        mimetypes.init()
        t, e = mimetypes.guess_type(name, strict=False)
        if t:
            return t

    def set_mimetype(self, mt):
        self._mimetype = mt

    mimetype = property(get_mimetype, set_mimetype)

    def set_values(self, page_data):
        "Only values allowed: status, headers, mimetype, filepath, enable_cache"
        if 'filepath' in page_data:
            self.filepath = page_data['filepath']
        if 'mimetype' in page_data:
            self._mimetype = page_data['mimetype']
        if 'enable_cache' in page_data:
            self.enable_cache = bool(page_data['enable_cache'])
        if ('headers' in page_data) and page_data['headers']:
            # headers have not been auto set, they are set specifically by the user
            self._headers_flag = False
        ParentPage.set_values(self, page_data)


    def update(self, environ, call_data, lang, ident_list=[]):
        """"If filepath set, then this is the file returned, if not, then projectfiles/project/static/name is returned"""
        self._environ = environ
        mimetype = self.mimetype
        if mimetype and self._headers_flag:
            # only add mimetype if headers auto set, not if headers specified in page_data
            self.headers.append(('content-type', mimetype))
        if not self.filepath:
            raise ServerError(message="Filepath not set")
        self._filepath_relative_to_project_files = os.path.join(skiboot.projectfiles(self.proj_ident), self.filepath)
        if not os.path.isfile(self._filepath_relative_to_project_files):
            # no need to do anything further
            return
        # get length of file
        if self._headers_flag:
            # only add content-length if headers auto set, not if headers specified in page_data
            self.headers.append(('content-length', str(os.path.getsize(self._filepath_relative_to_project_files))))
        # if a session cookie is specified, add it even if headers have been user set
        if self.session_cookie:
            self.headers.append(self.session_cookie)
        if self.language_cookie:
            self.headers.append(self.language_cookie)


    def _readfile(self, size=32768):
        "Return a generator reading the file"
        with open(self._filepath_relative_to_project_files, "rb") as f:
            data = f.read(size)
            while data:
                yield data
                data = f.read(size)


    def data(self):
        "returns an iterator reading the file"
        if not os.path.isfile(self._filepath_relative_to_project_files):
            # no need to do anything further
            return
        try:
            size = 32768
            if 'wsgi.file_wrapper' in self._environ:
                f = open(self._filepath_relative_to_project_files, "rb")
                return self._environ['wsgi.file_wrapper'](f, size)
            else:
                return self._readfile(size)
        except IOError:            
            return ["<!DOCTYPE HTML>\n<html>ERROR:UNABLE TO OPEN FILE\n</html>".encode('ascii', 'xmlcharrefreplace')]


    def __str__(self):
        return self.filepath

    def __repr__(self):
        if self.ident:
            return "File Page ident %s" % (self.ident,)
        else:
            return "File Page name %s" % (self.name,)


class CSS(ParentPage):
    """A CSS object represents a css file
    """


    def __init__(self, name="", brief="New CSS Page", style=collections.OrderedDict(), enable_cache=False):
        "style held as ordered dictionary each value being a list of lists"
        ParentPage.__init__(self, name=name, brief=brief)

        # self.style is a dictionary with keys being css selectors
        # and values being a list of two element lists
        # acting as css declaration blocks.

        self.style = style
        self.colour_substitution = {}
        # enable_cache flag
        self._enable_cache = False
        self.enable_cache = enable_cache
        # imports
        self.imports = []

    def set_enable_cache(self, enable_cache):
        "Sets enable cache in header"
        if enable_cache:
            self._enable_cache = True
            self.headers = [('content-type', 'text/css'), ('cache-control', 'max-age=3600')]
        else:
            self._enable_cache = False
            self.headers = [('content-type', 'text/css'),
                            ('cache-control','no-cache, no-store, must-revalidate'),
                            ('Pragma', 'no-cache'),
                            ( 'Expires', '0')]

    def get_enable_cache(self):
        return self._enable_cache

    enable_cache = property(get_enable_cache, set_enable_cache)


    def set_values(self, page_data):
        """enable_cache, @import, status, headers and ident_data"""
        if 'enable_cache' in page_data:
            self.enable_cache = bool(page_data['enable_cache'])
            del page_data['enable_cache']
        if 'colour_substitution' in page_data:
            self.colour_substitution = page_data['colour_substitution']
            del page_data['colour_substitution']
        if '@import' in page_data:
            if isinstance(page_data['@import'], list):
                self.imports = page_data['@import']
            elif isinstance(page_data['@import'], str):
                self.imports = [page_data['@import']]
            del page_data['@import']
        ParentPage.set_values(self, page_data)


    def selector_list(self):
        return list(self._style.keys())

    def selector_properties(self, selector):
        "Returns the property strings of the selector"
        style_text = ""
        if selector not in self._style:
            return ""
        value = self._style[selector]
        if value:
            # value is a list of two element lists
            # created value_list, which is value, sorted by the first element in each sub list
            value_list = sorted(value, key=lambda val : val[0])
            for a,b in value_list:
                style_text += "{a} : {b};\n".format(a=a, b=b)
        return style_text

    # property style is a dictionary of lists
    def get_style(self):
        return self._style

    def set_style(self, style):
        if isinstance(style, collections.OrderedDict):
            self._style = style
        elif isinstance(style, dict):
            self._style = collections.OrderedDict([(selector,value) for selector,value in style.items()])
        else:
            self._style = collections.OrderedDict()

    style = property(get_style, set_style)

    def write(self, filepath):
        "Writes a css file to the given filepath"
        if not self._style:
            return
        with open(filepath, "w") as f:
            f.write(self.__str__())

    def data(self):
        "Returns the page as a list of binary strings"
        if not self._style:
            return []
        style_binary = ["@charset \"UTF-8\";\n".encode('UTF-8')]
        if self.imports:
            for imp in self.imports:
                style_binary.append("@import : {imp};\n".format(imp=imp.strip(';')).encode('UTF-8'))
        for selector,value in self._style.items():
            if value:
                style_binary.append("\n{selector} {{".format(selector=selector).encode('UTF-8'))
                # value is a list of two element lists
                # created value_list, which is value, sorted by the first element in each sub list
                value_list = sorted(value, key=lambda val : val[0])
                for a,b in value_list:
                    if self.colour_substitution and ('$' in b):
                        c = Template(b)
                        b= c.safe_substitute(self.colour_substitution)
                    style_binary.append("\n{a} : {b};".format(a=a, b=b).encode('UTF-8'))
                style_binary.append("}\n".encode('UTF-8'))
        return style_binary

    def __str__(self):
        "Returns the css page as a string"
        if not self._style:
            return ""
        style_text = ""
        if self.imports:
            for imp in self.imports:
                style_text += "@import : {imp};\n".format(imp=imp.strip(';'))
        for selector, value in self._style.items():
            if value:
                style_text += """
{selector} {{""".format(selector=selector)
                # value is a list of two element lists
                # created value_list, which is value, sorted by the first element in each sub list
                value_list = sorted(value, key=lambda val : val[0])
                for a,b in value_list:
                    style_text += "\n{a} : {b};".format(a=a, b=b)
                style_text += "}\n"
        return style_text

    def __repr__(self):
        if self.ident:
            return "CSS Page ident %s" % (self.ident,)
        else:
            return "CSS Page name %s" % (self.name,)


class JSON(ParentPage):
    """A JSON object represents dynamically created json data
    """

    def __init__(self, name="", brief="New JSON Page", content=None, enable_cache=False):
        "content held as ordered dictionary"
        ParentPage.__init__(self, name=name, brief=brief)
        if content:
            self.content = content
        else:
            self.content = collections.OrderedDict()
        # enable_cache flag
        self._enable_cache = False
        self.enable_cache = enable_cache
        # setting jsondict overrides self.content
        self.jsondict = None

    def set_enable_cache(self, enable_cache):
        "Sets enable cache in header"
        if enable_cache:
            self._enable_cache = True
            self.headers = [('content-type', 'application/json'), ('cache-control', 'max-age=3600')]
        else:
            self._enable_cache = False
            self.headers = [('content-type', 'application/json'),
                                             ('cache-control','no-cache, no-store, must-revalidate'),
                                              ('Pragma', 'no-cache'),
                                              ( 'Expires', '0')]

    def get_enable_cache(self):
        return self._enable_cache

    enable_cache = property(get_enable_cache, set_enable_cache)

    def add_widgfield(self, widgfield, value):
        "Add widgfield and value to self.content"
        # ensure widgfield is a skiboot.WidgField object
        fld = skiboot.make_widgfield(widgfield)
        if not fld:
            return
        self.content[str(fld)] = value

    def del_widgfield(self, widgfield):
        "Removes widgfield from self.content"
        # ensure widgfield is a skiboot.WidgField object
        fld = skiboot.make_widgfield(widgfield)
        if not fld:
            return
        strfld = str(fld)
        if strfld in self.content:
            del self.content[strfld]

    def set_values(self, page_data):
        """Sets json content"""
        if not page_data:
            return
        if 'ClearAllErrors' in page_data:
            self.content["ClearAllErrors"] = bool(page_data['ClearAllErrors'])
            del page_data['ClearAllErrors']
        if 'localStorage' in page_data:
            if not isinstance(page_data['localStorage'], dict):
                raise ServerError("localStorage must be a dictionary")
            self.content["localStorage"] = page_data['localStorage']
            del page_data['localStorage']
        if 'sessionStorage' in page_data:
            if not isinstance(page_data['sessionStorage'], dict):
                raise ServerError("sessionStorage must be a dictionary")
            self.content["sessionStorage"] = page_data['sessionStorage']
            del page_data['sessionStorage']
        if 'throw' in page_data:
            self.content["throw"] = page_data['throw']
            del page_data['throw']
        if "IntervalTarget" in page_data:
            url = skiboot.get_url(page_data["IntervalTarget"], proj_ident=self.proj_ident)
            if url:
                self.content["IntervalTarget"] = url
            del page_data["IntervalTarget"]
        if 'interval' in page_data:
            interval=0
            try:
                interval = page_data['interval']
                del page_data['interval']
                interval = int(interval)
            except:
                pass
            else:
                self.content["interval"]=interval
        if "CatchToHTML" in page_data:
            url = skiboot.get_url(page_data["CatchToHTML"], proj_ident=self.proj_ident)
            if url:
                self.content["CatchToHTML"] = url
            del page_data["CatchToHTML"]
        if "JSONtoHTML" in page_data:
            url = skiboot.get_url(page_data["JSONtoHTML"], proj_ident=self.proj_ident)
            if url:
                self.content["JSONtoHTML"] = url
                return
            del page_data["JSONtoHTML"]
        ParentPage.set_values(self, page_data)
        for field, item in page_data.items():
            if field == 'lang':
                if isinstance(item, 'tuple') or isinstance(item, 'list'):
                    self.content['lang'] = item[0]
                else:
                    self.content['lang'] = item
            elif isinstance(field, str):
                self.content[field] = item
            else:
                widgfield = str(skiboot.make_widgfield(field))
                self.content[widgfield] = item

    def update(self, environ, call_data, lang, ident_list=[]):
        "Adds session_cookie, and also ident_list if debug mode is on"
        if self.session_cookie:
            self.headers.append(self.session_cookie)
        if self.language_cookie:
            self.headers.append(self.language_cookie)
        if skiboot.get_debug():
            # include environ, call_data and ident_list in json file
            self.content['environ'] = pprint.pformat(environ)
            if call_data:
                self.content['call_data'] = pprint.pformat(call_data)
            if not ident_list:
                self.content['ident_list'] = [[self.ident.to_comma_str(), 'This page', self.brief]]
                return
            idents = []
            for ident in ident_list:
                item = skiboot.from_ident(ident)
                idents.append([ident.to_comma_str(), item.responder.__class__.__name__, item.brief])
            idents.append([self.ident.to_comma_str(), 'This page', self.brief])
            self.content['ident_list'] = idents

    def show_error(self, error_messages=None):
        """Sets show error in page data"""
        if not error_messages:
            # send a default page error
            self.content["show_error"] = ""
            return
        for e in error_messages:
            # e is an instance of ErrorMessage
            if e.widget:
                displaywidget = str(skiboot.WidgField(s=e.section, w=e.widget, f='show_error', i=''))
            else:
                displaywidget = "show_error"
            self.content[displaywidget] = e.message

    def data(self):
        "Returns the page as a list of binary strings"
        if self.jsondict is not None:
            return [json.dumps(self.jsondict).encode('UTF-8')]
        if self.ident_data:
            # encode the ident data as base64
            b64binarydata = urlsafe_b64encode(self.ident_data.encode('ascii')).rstrip(b"=")  # removes final '=' padding
            self.content['ident_data'] = b64binarydata.decode('ascii')
        elif 'ident_data' in self.content:
            del self.content['ident_data']
        return [json.dumps(self.content).encode('UTF-8')]

    def __str__(self):
        "Returns the json page as a string"
        if self.jsondict is not None:
            return json.dumps(self.jsondict)
        if self.ident_data:
            # encode the ident data as base64
            b64binarydata = urlsafe_b64encode(self.ident_data.encode('ascii')).rstrip(b"=")  # removes final '=' padding
            self.content['ident_data'] = b64binarydata.decode('ascii')
        elif 'ident_data' in self.content:
            del self.content['ident_data']
        return json.dumps(self.content)

    def __repr__(self):
        if self.ident:
            return "JSON Page ident %s" % (self.ident,)
        else:
            return "JSON Page name %s" % (self.name,)





# This skis project is added to other projects and consists of
# javascript and css files. As it is essentially a 'library'
# it is usually given a URL of /lib when added - but this is
# not a necessity, just convention.

import os
from . import WSGIApplication

PROJECTFILES = os.path.dirname(os.path.realpath(__file__))

def start_call(called_ident, skicall):
    "This project has no start_call functionality"
    return called_ident

def submit_data(skicall):
    "This project has no submit_data functionality"
    return

def end_call(page_ident, page_type, skicall):
    "This project has no end_call functionality"
    return

# As this project is not intended to run as a stand-alone service, a function
# is provided rather than an application object being immediately created.

def makeapp():
    """This function returns the skis application."""

    # The WSGIApplication created here is generally given a URL of "/lib"
    # when added to the root project using application.add_project

    return WSGIApplication('skis', PROJECTFILES, {}, start_call, submit_data, end_call)



############
#
#  To use this module - in your own code you would normally create your own 'application' object
#  and then call this module's makeapp function to produce a 'skis_application'
#  You would then add skis_application to your application using
#  application.add_project(skis_application, url='/lib')
#
#  from skipole import skis
#  skis_application = skis.makeapp()
#  application.add_project(skis_application, url='/lib')
#
############






"Functions implementing part editing"

import json

from skipole import skilift
from skipole.skilift import fromjson, editpage, editsection, item_info
from skipole import FailPage, ValidateError, GoTo, ServerError



def retrieve_editpart(skicall):
    "Fills in the edit a part page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # a skilift.part_tuple is (project, pagenumber, page_part, section_name, name, location, part_type, brief)

    pagenumber = None
    section_name = None

    if 'part_tuple' in call_data:
        part_tuple = call_data['part_tuple']
        project = part_tuple.project
        location = part_tuple.location
        pagenumber = part_tuple.pagenumber
        section_name = part_tuple.section_name
    else:
        project = call_data['editedprojname']
        location = call_data['location']
        if 'page_number' in call_data:
            pagenumber = call_data['page_number']
        else:
            section_name = call_data['section_name']

    if (pagenumber is None) and (section_name is None):
        raise ValidateError("Page/section not identified")

    call_data['location'] = location

    part = None

    try:
        if pagenumber:
            call_data['page_number'] = pagenumber
            pchange = call_data['pchange']
            part = editpage.page_element(project, pagenumber, pchange, location)
        else:
            call_data['section_name'] = section_name
            schange = call_data['schange']
            part = editsection.section_element(project, section_name, schange, location)

        if part is None:
            raise FailPage("Part not identified")
    except ServerError as e:
        raise FailPage(e.message)

    # Fill in header
    page_data[("adminhead","page_head","large_text")] = "Edit the html element."

    # header done, now page contents

    if (part.part_type == "Part") or (part.part_type == "Section"):
        ending = ">"
    else:
        ending = " />"

    atts_list = []
    vals_list = []
    if part.attribs:
        for att in part.attribs:
            atts_list.append(att)
        # sort atts_list
        atts_list.sort()
        attstring = ""
        for att in atts_list:
            vals_list.append(part.attribs[att])
            attstring += " %s = \"%s\"" % (att, part.attribs[att])
        page_data[('tag_para','para_text')] = "Element tag : <%s%s%s" % (part.tag_name, attstring, ending)
    else:
        page_data[('tag_para','para_text')] = "Element tag : <%s%s" % (part.tag_name, ending)

    # list table of attributes, widget tables.Table2_1
    # contents = list of lists, each inner has five elements
    # col 0 is the text to place in the first column,
    # col 1 is the text to place in the second column,
    # col 2, 3, 4 are the three get field contents of the link
    if len(atts_list):
        page_data[('attribs_list_para','show')] = True
        page_data[('attribs_list','show')] = True
        contents = []
        for index, att in enumerate(atts_list):
            row = [ att, vals_list[index], att]
            contents.append(row)
        page_data['attribs_list', 'contents'] = contents
    else:
        page_data[('attribs_list_para','show')] = False
        page_data[('attribs_list','show')] = False

    # input form to change the tag name
    page_data[('tag_input','input_text')] = part.tag_name

    # input form to change the tag brief
    page_data[('tag_brief','input_text')] = part.brief

    if part.part_type == "Part":
        if part.hide_if_empty:
            page_data[('hidecheck','checked')] = True
        else:
            page_data[('hidecheck','checked')] = False
    else:
        page_data[('set_to_hide','show')] = False
        page_data[('para_to_hide','show')] = False
        page_data[('para_to_hide2','show')] = False
        page_data['partdownload', 'show'] = False
        page_data['aboutdownload', 'show'] = False


def set_tag(skicall):
    "Sets the part tag name, or brief, adds an attribute"

    call_data = skicall.call_data
    page_data = skicall.page_data

    pagenumber = None
    section_name = None

    project = call_data['editedprojname']
    location = call_data['location']
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
        pchange = call_data['pchange']
    else:
        section_name = call_data['section_name']
        schange = call_data['schange']

    if (pagenumber is None) and (section_name is None):
        raise ValidateError("Page/section not identified")

    try:
        if pagenumber:
            part = editpage.page_element(project, pagenumber, pchange, location)
        else:
            part = editsection.section_element(project, section_name, schange, location)

        if part is None:
            raise FailPage("Part not identified")

        message = "A new value to edit has not been found"

        if 'tag_name' in call_data:
            tag_name = call_data["tag_name"]
            message = 'New tag set'
        else:
            tag_name = part.tag_name

        if 'tag_brief' in call_data:
            brief = call_data["tag_brief"]
            message = 'New description set'
        else:
            brief = part.brief

        attribs = part.attribs

        if 'attrib' in call_data:
            if 'val' not in call_data:
                raise FailPage("The attribute value has not been found")
            attribs.update({call_data["attrib"]:call_data["val"]})
            message = 'Attribute updated'

        if 'hide_if_empty' in call_data:
            hide = call_data['hide_if_empty']
            if hide == 'hide':
                hide_if_empty = True
                message = 'Element will be hidden if no content within it.'
            else:
                hide_if_empty = False
                message = 'Element will be shown even if no content within it.'
        else:
            hide_if_empty = part.hide_if_empty

        if pagenumber:
            call_data['pchange'] = editpage.edit_page_element(project, pagenumber, pchange, location, tag_name, brief, hide_if_empty, attribs)
        else:
            call_data['schange'] = editsection.edit_section_element(project, section_name, schange, location, tag_name, brief, hide_if_empty, attribs)

    except ServerError as e:
        raise FailPage(e.message)

    call_data['status'] = message


def remove_tag_attribute(skicall):
    "Removes the given tag attribute"

    call_data = skicall.call_data
    page_data = skicall.page_data

    pagenumber = None
    section_name = None

    project = call_data['editedprojname']
    location = call_data['location']
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
        pchange = call_data['pchange']
    else:
        section_name = call_data['section_name']
        schange = call_data['schange']

    if (pagenumber is None) and (section_name is None):
        raise ValidateError("Page/section not identified")

    if ('attribs_list','contents') not in call_data:
        raise FailPage("A Tag attribute to remove has not been found")

    att = call_data['attribs_list','contents']

    try:
        if pagenumber:
            part = editpage.page_element(project, pagenumber, pchange, location)
        else:
            part = editsection.section_element(project, section_name, schange, location)

        if part is None:
            raise FailPage("Part not identified")

        if part.attribs and (att in part.attribs):
            # set the attribute to be removed in the input boxes so it can be edited
            page_data['add_attrib', 'input_text1'] = att
            page_data['add_attrib', 'input_text2'] = part.attribs[att]
        else:
            raise FailPage("A Tag attribute to remove has not been found")

        if pagenumber:
            call_data['pchange'] = editpage.del_attrib(project, pagenumber, pchange, location, att)
        else:
            call_data['schange'] = editsection.del_attrib(project, section_name, schange, location, att)

    except ServerError as e:
        raise FailPage(e.message)




def downloadpart(skicall):
    "Gets part, and returns a json dictionary, this will be sent as an octet file to be downloaded"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        pagenumber = None

    if 'section_name' in call_data:
        section_name = call_data['section_name']
    else:
        section_name = None   

    part_info = skilift.part_info(project, pagenumber, section_name, call_data['location'])

    if not part_info:
        raise FailPage("Part not identified")

    parttext, part_dict = fromjson.item_outline(project, pagenumber, section_name, call_data['location'])
    # set version and skipole as the first two items in the dictionary
    versions = skilift.versions(project)
    part_dict["skipole"] = versions.skipole
    part_dict.move_to_end('skipole', last=False)
    part_dict["version"] = versions.project
    part_dict.move_to_end('version', last=False)
    jsonstring = json.dumps(part_dict, indent=4, separators=(',', ':'))
    line_list = []
    n = 0
    for line in jsonstring.splitlines(True):
        binline = line.encode('utf-8')
        n += len(binline)
        line_list.append(binline)
    page_data['headers'] = [('content-type', 'application/octet-stream'), ('content-length', str(n))]
    return line_list


def create_insert(skicall):
    "Creates the html element"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    try:
        part_top, container, location_integers = call_data['location']
        if call_data['newopenclosed'] == 'open':
            opentag = True
        else:
            opentag = False

        if 'page_number' in call_data:
            pagenumber = call_data['page_number']
            page_info = item_info(project, pagenumber)
            if page_info is None:
                raise FailPage("Page to edit not identified")

            if (page_info.item_type != "TemplatePage") and (page_info.item_type != "SVG"):
                raise FailPage("Page not identified")

            # page to go back to
            target = None
            if part_top == 'head':
                target = "page_head"
            elif part_top == 'body':
                target = "page_body"
            elif part_top == 'svg':
                target = "page_svg"

            if container is not None:
                target = "back_to_container"

            if target is None:
                raise FailPage("Invalid location")

            call_data['pchange'] = editpage.create_html_element_in_page(project,
                                                                        pagenumber,
                                                                        call_data['pchange'],
                                                                        call_data['location'],
                                                                        call_data['newpartname'],
                                                                        call_data['newbrief'],
                                                                        opentag)

        elif 'section_name' in call_data:
            section_name = call_data['section_name']

            if container is not None:
                target = "back_to_container"
            else:
                target = "back_to_section"

            call_data['schange'] = editsection.create_html_element_in_section(project,
                                                                              section_name,
                                                                              call_data['schange'],
                                                                              call_data['location'],
                                                                              call_data['newpartname'],
                                                                              call_data['newbrief'],
                                                                              opentag)

        else:
            raise FailPage("Either a page or section must be specified")

    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = 'New tag inserted'
    raise GoTo(target = target, clear_submitted=True)







"""Contains form widgets, these have 'container' functionality - they can contain further html and widgets, typically
      further input fields.  The module also has an Hidden Field and Submit Button widgets, which can be inserted into
     a form. """

from .. import skiboot, tag, excepts
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class HiddenField(ClosedWidget):
    """An input field of type hidden, for use as an insert into form widgets"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'hidden_field':FieldArg("text", '', valdt=True, jsonset=True)}

    def __init__(self, name=None, brief='', **field_args):
        "hidden_field: A hidden input field"
        ClosedWidget.__init__(self, name=name, tag_name="input", brief=brief, **field_args)

    def _build(self, page, ident_list, environ, call_data, lang):
        "Sets the attributes"
        value = self.get_field_value('hidden_field')
        if not value:
            self.show = False
            return
        self.update_attribs({"name":self.get_formname('hidden_field'),
                       "value":value,
                       "type":"hidden"})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<input type="hidden" /> <!-- with widget id and class widget_class -->
  <!-- with value of the "hidden_field" value, and name being the widgfield -->"""


class HiddenSessionStorage(ClosedWidget):
    """An input field of type hidden, for use as an insert into form widgets"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'session_key':FieldArg("text", '', valdt=True, jsonset=True)}

    def __init__(self, name=None, brief='', **field_args):
        "hidden_field: A hidden input field with value from session storage"
        ClosedWidget.__init__(self, name=name, tag_name="input", brief=brief, **field_args)
        self._key = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        "Sets the attributes"
        self._key = self.get_field_value('session_key')
        if not self._key:
            self.show = False
            return
        self.update_attribs({"name":self.get_formname('session_key'),
                       "value":"",
                       "type":"hidden"})

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets key value into the value attribute by calling the widget updatefunc"""
        if not self._key:
            return
        jscript =  """
  SKIPOLE.widgets["{ident}"].updatefunc();
""".format(ident=self.get_id())
        return self._make_fieldvalues(session_key=self._key) + jscript

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<input type="hidden" /> <!-- with widget id and class widget_class -->
  <!-- with value taken from the session storage with key "session_key", and name being the session_key widgfield -->"""



class HiddenLocalStorage(ClosedWidget):
    """An input field of type hidden, for use as an insert into form widgets"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'local_key':FieldArg("text", '', valdt=True, jsonset=True)}

    def __init__(self, name=None, brief='', **field_args):
        "hidden_field: A hidden input field with value from local storage"
        ClosedWidget.__init__(self, name=name, tag_name="input", brief=brief, **field_args)
        self._key = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        "Sets the attributes"
        self._key = self.get_field_value('local_key')
        if not self._key:
            self.show = False
            return
        self.update_attribs({"name":self.get_formname('local_key'),
                       "value":"",
                       "type":"hidden"})

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets key value into the value attribute by calling the widget updatefunc"""
        if not self._key:
            return
        jscript = """
  SKIPOLE.widgets["{ident}"].updatefunc();
""".format(ident=self.get_id())
        return self._make_fieldvalues(local_key=self._key) + jscript



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<input type="hidden" /> <!-- with widget id and class widget_class -->
  <!-- with value taken from the local storage with key "local_key", and name being the local_key widgfield -->"""



class SubmitButton1(ClosedWidget):
    """An input field of type submit, for use as an insert into form widgets"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'button_text':FieldArg("text", 'Submit', valdt=True, jsonset=True)}

    def __init__(self, name=None, brief='', **field_args):
        "Create input type submit button widget"
        ClosedWidget.__init__(self, name=name, tag_name="input", brief=brief, **field_args)

    def _build(self, page, ident_list, environ, call_data, lang):
        "Sets the attributes"
        button_text = self.get_field_value('button_text')
        if not button_text:
            button_text = "Submit"
        self.update_attribs({"name":self.get_formname('button_text'),
                       "value":button_text,
                       "type":"submit"})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<input type="button_text" /> <!-- with widget id and class widget_class -->
  <!-- with value of the "button_text" value, and name being the 'button_text widgfield -->"""


class SubmitButton2(ClosedWidget):
    """An input field of type submit, for use as an insert into form widgets"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'button_text':FieldArg("text", 'Submit', jsonset=True)}

    def __init__(self, name=None, brief='', **field_args):
        "Create input type submit button widget"
        ClosedWidget.__init__(self, name=name, tag_name="input", brief=brief, **field_args)

    def _build(self, page, ident_list, environ, call_data, lang):
        "Sets the attributes"
        button_text = self.get_field_value('button_text')
        if not button_text:
            button_text = "Submit"
        self.update_attribs({"value":button_text, "type":"submit"})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<input type="button_text" /> <!-- with widget id and class widget_class -->
  <!-- with value of the "button_text" value, but no name, so does not submit a widgfield -->"""


class Form1(Widget):
    """A form with a container and four hidden fields. Used with further input fields set within it.
       On error - the error message is displayed before any of the contents
       Does not include a submit button, therefore requires one to be inserted with the contents"""

    _container = ((1,0),)

    error_location = (0,0,0)

    arg_descriptions = {'action':FieldArg("url", ''),
                        'enctype':FieldArg("text", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True),
                        'container_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", '')
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        action: The page ident, label, url this form links to
        enctype: Sets the enctype attribute if given
        hidden_field1: A hidden field value, leave blank if unused
        hidden_field2: A second hidden field value, leave blank if unused
        hidden_field3: A third hidden field value, leave blank if unused
        hidden_field4: A fourth hidden field value, leave blank if unused
        container_class: the class attribute of the div holding the container
        error_class: The class applied to the paragraph containing the error message on error."""
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self.update_attribs({"role":"form", "method":"post"})
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})
        # The location 1,0 is available as a container
        self[1][0] = tag.Part(tag_name="div")
        self[1][0][0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if not self.get_field_value("action"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = skiboot.get_url(self.get_field_value("action"),  proj_ident=page.proj_ident)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[1].update_attribs({"action": actionurl})
        if self.get_field_value('enctype'):
            self[1].update_attribs({"enctype": self.get_field_value('enctype')})

        # the div holding the container
        if self.get_field_value('container_class'):
            self[1][0].attribs = {"class": self.get_field_value('container_class')}

        # add ident and four hidden fields
        self.add_hiddens(self[1], page)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        jscript = """  $('#{ident} form').on("submit", function(e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident=self.get_id())
        return jscript

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <form method=\"post\"> <!-- action attribute set to action field -->
    <div> <!-- this div has the class attribute set to container_class -->
      <!-- container 0 for further html -->
    </div>
    <!-- hidden input fields -->                              
  </form>
</div>"""


class SubmitForm1(Widget):
    """A form taking contents with submit button, left or right labels and four hidden fields.
       Used with further input fields set within it. On error - the error message is displayed
       below the form tag, before any of the contents"""

    _container = ((1,0),)

    error_location = (0,0,0)

    arg_descriptions = {'left_label':FieldArg("text", 'Please Submit:'),
                        'left_class':FieldArg("cssclass", ''),
                        'left_style':FieldArg("cssstyle", ''),
                        'right_label':FieldArg("text", ''),
                        'right_class':FieldArg("cssclass", ''),
                        'right_style':FieldArg("cssstyle", ''),
                        'action_json':FieldArg("url", ''),
                        'action':FieldArg("url", ''),
                        'enctype':FieldArg("text", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True),
                        'button_text':FieldArg("text",'Submit'),
                        'button_wait_text':FieldArg("text", ''),
                        'button_class':FieldArg("cssclass", ''),
                        'div_class':FieldArg("cssclass", ''),
                        'container_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", ''),
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        left_label: The text displayed to the left of the button
        left_class: The css class of the label to the left of the button
        right_label: The text displayed to the right of the button
        right_class: The css class of the label to the right of the button
        action_json:  if a value set, and client has jscript enabled, this is the page ident, label, url this button links to, expects a json page back
        action: The page ident, label, url this button links to
        enctype: Sets the enctype attribute if given
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        button_text: The text on the button
        button_wait_text: A 'please wait' message shown on the button
        button_class: The css class of the button
        div_class: the class attribute of the div tag which contains the label and button
        container_class: the class attribute of the div holding the container
        error_class: The class applied to the paragraph containing the error message on error.
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})
        # The location 1,0 is available as a container
        self[1][0] = tag.Part(tag_name='div')
        self[1][0][0] = ''
        # tag containing label and button
        self[1][1] = tag.Part(tag_name='div')
        # the left label
        self[1][1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the submit button
        self[1][1][1] = tag.ClosedPart(tag_name="input")
        # the right label
        self[1][1][2] = tag.Part(tag_name="label", hide_if_empty=True)
        self._jsonurl = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        self._jsonurl = skiboot.get_url(self.get_field_value("action_json"), proj_ident=page.proj_ident)
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if not self.get_field_value("action"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = skiboot.get_url(self.get_field_value("action"),  proj_ident=page.proj_ident)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[1].update_attribs({"action": actionurl})
        if self.get_field_value('enctype'):
            self[1].update_attribs({"enctype": self.get_field_value('enctype')})

        # the div holding the container
        if self.get_field_value('container_class'):
            self[1][0].attribs = {"class": self.get_field_value('container_class')}

        # the div holding label and button
        if self.get_field_value('div_class'):
            self[1][1].attribs = {"class": self.get_field_value('div_class')}

        if self.get_field_value('left_label'):
            self[1][1][0][0] = self.get_field_value('left_label')
        if self.get_field_value('left_class'):
            self[1][1][0].attribs = {"class": self.get_field_value('left_class')}
        if self.get_field_value('left_style'):
            self[1][1][0].attribs = {"style": self.get_field_value('left_style')}

        # submit button
        if self.get_field_value('button_class'):
            self[1][1][1].attribs = {"value":self.get_field_value('button_text'), "type":"submit", "class": self.get_field_value('button_class')}
        else:
            self[1][1][1].attribs = {"value":self.get_field_value('button_text'), "type":"submit"}

        # set an id in the submit button for the 'label for' tag
        self[1][1][1].insert_id()

        if self.get_field_value('right_label'):
            self[1][1][2][0] = self.get_field_value('right_label')
        if self.get_field_value('right_class'):
            self[1][1][2].attribs = {"class": self.get_field_value('right_class')}
        if self.get_field_value('right_style'):
            self[1][1][2].attribs = {"style": self.get_field_value('right_style')}

        # set the label 'for' attribute
        self[1][1][0].update_attribs({'for':self[1][1][1].get_id()})
        self[1][1][2].update_attribs({'for':self[1][1][1].get_id()})

        # add ident and four hidden fields
        self.add_hiddens(self[1], page)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        jscript = """$('#{ident} form').on("submit", function(e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident=self.get_id())
        if self._jsonurl:
            return jscript + self._make_fieldvalues('button_wait_text', buttonident = self[1][1][1].get_id(), url=self._jsonurl)
        else:
            return jscript + self._make_fieldvalues('button_wait_text', buttonident = self[1][1][1].get_id())


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <form method="post"> <!-- action attribute set to action field -->
    <div>  <!-- this div has the class attribute set to container_class -->
      <!-- container 0 for further html -->
    </div>
    <div>  <!-- this div has the class attribute set to div_class -->
      <label> <!-- with class set to left_class and content to left_label -->
      </label>
      <input type=\"submit\" /> <!-- button value set to button_text -->
      <label> <!-- with class set to right_class and content to right_label -->
      </label>
    </div>
    <!-- hidden input fields -->                              
  </form>
</div>"""



class SubmitForm2(Widget):
    """A form taking contents with submit button, left or right labels and four hidden fields.
       Used with further input fields set within it. On error - the error message is displayed
       below the form tag, before any of the contents
       Can send session or local storage values."""

    _container = ((1,0),)

    error_location = (0,0,0)

    arg_descriptions = {'left_label':FieldArg("text", 'Please Submit:'),
                        'left_class':FieldArg("cssclass", ''),
                        'left_style':FieldArg("cssstyle", ''),
                        'right_label':FieldArg("text", ''),
                        'right_class':FieldArg("cssclass", ''),
                        'right_style':FieldArg("cssstyle", ''),
                        'action_json':FieldArg("url", ''),
                        'action':FieldArg("url", ''),
                        'enctype':FieldArg("text", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True),
                        'session_storage':FieldArg("text", "", valdt=True, jsonset=True),
                        'local_storage':FieldArg("text","", valdt=True, jsonset=True),
                        'button_text':FieldArg("text",'Submit'),
                        'button_wait_text':FieldArg("text", ''),
                        'button_class':FieldArg("cssclass", ''),
                        'div_class':FieldArg("cssclass", ''),
                        'container_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", ''),
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        left_label: The text displayed to the left of the button
        left_class: The css class of the label to the left of the button
        right_label: The text displayed to the right of the button
        right_class: The css class of the label to the right of the button
        action_json:  if a value set, and client has jscript enabled, this is the page ident, label, url this button links to, expects a json page back
        action: The page ident, label, url this button links to
        enctype: Sets the enctype attribute if given
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        session_storage: A session storage key, this widgfield returns the stored value if anything
        local_storage: A local storage key, this widgfield returns the stored value if anything
        button_text: The text on the button
        button_wait_text: A 'please wait' message shown on the button
        button_class: The css class of the button
        div_class: the class attribute of the div tag which contains the label and button
        container_class: the class attribute of the div holding the container
        error_class: The class applied to the paragraph containing the error message on error.
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})
        # The location 1,0 is available as a container
        self[1][0] = tag.Part(tag_name='div')
        self[1][0][0] = ''
        # tag containing label and button
        self[1][1] = tag.Part(tag_name='div')
        # the left label
        self[1][1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the submit button
        self[1][1][1] = tag.ClosedPart(tag_name="input")
        # the right label
        self[1][1][2] = tag.Part(tag_name="label", hide_if_empty=True)
        self._jsonurl = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        self._jsonurl = skiboot.get_url(self.get_field_value("action_json"), proj_ident=page.proj_ident)
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if not self.get_field_value("action"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = skiboot.get_url(self.get_field_value("action"),  proj_ident=page.proj_ident)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[1].update_attribs({"action": actionurl})
        if self.get_field_value('enctype'):
            self[1].update_attribs({"enctype": self.get_field_value('enctype')})

        # the div holding the container
        if self.get_field_value('container_class'):
            self[1][0].attribs = {"class": self.get_field_value('container_class')}

        # the div holding label and button
        if self.get_field_value('div_class'):
            self[1][1].attribs = {"class": self.get_field_value('div_class')}

        if self.get_field_value('left_label'):
            self[1][1][0][0] = self.get_field_value('left_label')
        if self.get_field_value('left_class'):
            self[1][1][0].attribs = {"class": self.get_field_value('left_class')}
        if self.get_field_value('left_style'):
            self[1][1][0].attribs = {"style": self.get_field_value('left_style')}

        # submit button
        if self.get_field_value('button_class'):
            self[1][1][1].attribs = {"value":self.get_field_value('button_text'), "type":"submit", "class": self.get_field_value('button_class')}
        else:
            self[1][1][1].attribs = {"value":self.get_field_value('button_text'), "type":"submit"}

        # set an id in the submit button for the 'label for' tag
        self[1][1][1].insert_id()

        if self.get_field_value('right_label'):
            self[1][1][2][0] = self.get_field_value('right_label')
        if self.get_field_value('right_class'):
            self[1][1][2].attribs = {"class": self.get_field_value('right_class')}
        if self.get_field_value('right_style'):
            self[1][1][2].attribs = {"style": self.get_field_value('right_style')}

        # set the label 'for' attribute
        self[1][1][0].update_attribs({'for':self[1][1][1].get_id()})
        self[1][1][2].update_attribs({'for':self[1][1][1].get_id()})

        # add ident and four hidden fields
        self.add_hiddens(self[1], page)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        jscript = """$('#{ident} form').on("submit", function(e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident=self.get_id())
        if self._jsonurl:
            return jscript + self._make_fieldvalues('button_wait_text',
                                                    'session_storage',
                                                    'local_storage',
                                                    buttonident = self[1][1][1].get_id(),
                                                    url=self._jsonurl)
        else:
            return jscript + self._make_fieldvalues('button_wait_text',
                                                    'session_storage',
                                                    'local_storage',
                                                    buttonident = self[1][1][1].get_id())


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <form method="post"> <!-- action attribute set to action field -->
    <div>  <!-- this div has the class attribute set to container_class -->
      <!-- container 0 for further html -->
    </div>
    <div>  <!-- this div has the class attribute set to div_class -->
      <label> <!-- with class set to left_class and content to left_label -->
      </label>
      <input type=\"submit\" /> <!-- button value set to button_text -->
      <label> <!-- with class set to right_class and content to right_label -->
      </label>
    </div>
    <!-- hidden input fields -->                              
  </form>
</div>"""









"""Functions for editing a validator"""

import pkgutil, importlib, inspect


from ..ski import skiboot, widgets, validators
from ..ski.excepts import ServerError

from . import get_proj_page, get_proj_section

from .info_tuple import ValidatorInfo


def _get_section_val_list(project, section_name, schange, widget_name, field_arg):
    "Returns proj, section, val_list"
    proj, section = get_proj_section(project, section_name, schange)
    widget = section.widgets.get(widget_name)
    if (not isinstance(widget, widgets.Widget)) and (not isinstance(widget, widgets.ClosedWidget)):
        raise ServerError("Widget not found")
    if field_arg not in widget.fields:
        raise ServerError("Field not found")
    field = widget.fields[field_arg]
    if not field.valdt:
        raise ServerError("Field does not take validators")
    return proj, section, field.val_list


def _get_page_val_list(project, pagenumber, pchange, widget_name, field_arg):
    "Returns proj, page, val_list"
    proj, page = get_proj_page(project, pagenumber, pchange)
    widget = page.widgets.get(widget_name)
    if (not isinstance(widget, widgets.Widget)) and (not isinstance(widget, widgets.ClosedWidget)):
        raise ServerError("Widget not found")
    if field_arg not in widget.fields:
        raise ServerError("Field not found")
    field = widget.fields[field_arg]
    if not field.valdt:
        raise ServerError("Field does not take validators")
    return proj, page, field.val_list


def validator_modules():
    "Returns a tuple of validator modules"
    return tuple(name for (module_loader, name, ispkg) in pkgutil.iter_modules(validators.__path__))


def validators_in_module(module_name):
    "Returns a tuple of validator names in a module"
    module = importlib.import_module("skipole.ski.validators." + module_name)
    return tuple(name for name,obj in inspect.getmembers(module, lambda member: inspect.isclass(member) and (member.__module__ == module.__name__)))


def get_section_field_validator_list(project, section_name, schange, widget_name, field_arg):
    "Returns list of validators attached to this widget field, each item in the list being a tuple of validator class name, module name)"
    proj, section, val_list = _get_section_val_list(project, section_name, schange, widget_name, field_arg)
    return tuple( (v.__class__.__name__, v.module_name()) for v in val_list )


def get_page_field_validator_list(project, pagenumber, pchange, widget_name, field_arg):
    "Returns list of validators attached to this widget field, each item in the list being a tuple of validator class name, module name)"
    proj, page, val_list = _get_page_val_list(project, pagenumber, pchange, widget_name, field_arg)
    return tuple( (v.__class__.__name__, v.module_name()) for v in val_list )


def page_field_validator_info(project, pagenumber, pchange, widget_name, field_arg, validx):
    "Returns a named tuple of info about a validator, validx is the index number of the validator within the validator list attached to the field"
    proj, page, val_list = _get_page_val_list(project, pagenumber, pchange, widget_name, field_arg)
    # get validator
    if val_list and (validx >= 0) and (validx < len(val_list)):
        validator = val_list[validx]
    else:
        raise ServerError("Unknown validator")
    return ValidatorInfo(validator.__class__.__name__,
                         validator.module_name(),
                         validator.message,
                         validator.message_ref,
                         validator.displaywidget.to_str_tuple(),
                         validator.allowed_values,
                         validator.val_args
                        )


def section_field_validator_info(project, section_name, schange, widget_name, field_arg, validx):
    "Returns a named tuple of info about a validator, validx is the index number of the validator within the validator list attached to the field"
    proj, section, val_list = _get_section_val_list(project, section_name, schange, widget_name, field_arg)
    # get validator
    if val_list and (validx >= 0) and (validx < len(val_list)):
        validator = val_list[validx]
    else:
        raise ServerError("Unknown validator")
    return ValidatorInfo(validator.__class__.__name__,
                         validator.module_name(),
                         validator.message,
                         validator.message_ref,
                         validator.displaywidget.to_str_tuple(),
                         validator.allowed_values,
                         validator.val_args
                        )
   

def create_section_field_validator(project, section_name, schange, widget_name, field_arg, validator_module, validator_name):
    "Add a validator to the field, return the new section change"
    proj, section = get_proj_section(project, section_name, schange)
    widget = section.widgets.get(widget_name)
    if (not isinstance(widget, widgets.Widget)) and (not isinstance(widget, widgets.ClosedWidget)):
        raise ServerError("Widget not found")
    if field_arg not in widget.fields:
        raise ServerError("Field not found")
    field = widget.fields[field_arg]
    if not field.valdt:
        raise ServerError("Field does not take validators")

    val_modules = validator_modules()
    if validator_module not in val_modules:
        raise ServerError("Validator module not found")
    module = importlib.import_module("skipole.ski.validators." + validator_module)

    validator_dict = {name:cls for (name,cls) in inspect.getmembers(module, lambda member: inspect.isclass(member) and (member.__module__ == module.__name__))}

    if validator_name not in validator_dict:
        raise ServerError("Validator not identified")

    validator_cls = validator_dict[validator_name]
    validator_instance = validator_cls()
    field.add_validator(validator_instance)

    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def create_page_field_validator(project, pagenumber, pchange, widget_name, field_arg, validator_module, validator_name):
    "Add a validator to the field, return the new page change"
    proj, page = get_proj_page(project, pagenumber, pchange)
    widget = page.widgets.get(widget_name)
    if (not isinstance(widget, widgets.Widget)) and (not isinstance(widget, widgets.ClosedWidget)):
        raise ServerError("Widget not found")
    if field_arg not in widget.fields:
        raise ServerError("Field not found")
    field = widget.fields[field_arg]
    if not field.valdt:
        raise ServerError("Field does not take validators")

    val_modules = validator_modules()
    if validator_module not in val_modules:
        raise ServerError("Validator module not found")
    module = importlib.import_module("skipole.ski.validators." + validator_module)

    validator_dict = {name:cls for (name,cls) in inspect.getmembers(module, lambda member: inspect.isclass(member) and (member.__module__ == module.__name__))}

    if validator_name not in validator_dict:
        raise ServerError("Validator not identified")

    validator_cls = validator_dict[validator_name]
    validator_instance = validator_cls()
    field.add_validator(validator_instance)

    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def remove_page_field_validator(project, pagenumber, pchange, widget_name, field_arg, validx):
    "Removes a validator, validx is the index number of the validator within the validator list attached to the field, return the new page change"
    proj, page, val_list = _get_page_val_list(project, pagenumber, pchange, widget_name, field_arg)
    # get validator
    if val_list and (validx >= 0) and (validx < len(val_list)):
        del val_list[validx]
    else:
        raise ServerError("Unknown validator")
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def remove_section_field_validator(project, section_name, schange, widget_name, field_arg, validx):
    "Removes a validator, validx is the index number of the validator within the validator list attached to the field, return the new section change"
    proj, section, val_list = _get_section_val_list(project, section_name, schange, widget_name, field_arg)
    # get validator
    if val_list and (validx >= 0) and (validx < len(val_list)):
        del val_list[validx]
    else:
        raise ServerError("Unknown validator")
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def swap_page_field_validators(project, pagenumber, pchange, widget_name, field_arg, validx1, validx2):
    "swaps validators at index positions validx1, validx2, which are positions within the validator list attached to the field, return the new page change"
    proj, page, val_list = _get_page_val_list(project, pagenumber, pchange, widget_name, field_arg)
    # swap validators
    try:
        val_list[validx1], val_list[validx2] = val_list[validx2], val_list[validx1]
    except Exception:
        raise ServerError("Invalid operation")
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def swap_section_field_validators(project, section_name, schange, widget_name, field_arg, validx1, validx2):
    "swaps validators at index positions validx1, validx2, which are positions within the validator list attached to the field, return the new section change"
    proj, section, val_list = _get_section_val_list(project, section_name, schange, widget_name, field_arg)
    # swap validators
    try:
        val_list[validx1], val_list[validx2] = val_list[validx2], val_list[validx1]
    except Exception:
        raise ServerError("Invalid operation")
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def set_section_field_validator_error_message(project, section_name, schange, widget_name, field_arg, validx, e_message):
    "Set the error message on the validator at index validx within the validator list attached to the field, return the new section change"
    proj, section, val_list = _get_section_val_list(project, section_name, schange, widget_name, field_arg)
    val_list[validx].message = e_message
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def set_page_field_validator_error_message(project, pagenumber, pchange, widget_name, field_arg, validx, e_message):
    "Set the error message on the validator at index validx within the validator list attached to the field, return the new page change"
    proj, page, val_list = _get_page_val_list(project, pagenumber, pchange, widget_name, field_arg)
    val_list[validx].message = e_message
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def set_section_field_validator_error_message_reference(project, section_name, schange, widget_name, field_arg, validx, e_message_ref):
    "Set the error message reference on the validator at index validx within the validator list attached to the field, return the new section change"
    proj, section, val_list = _get_section_val_list(project, section_name, schange, widget_name, field_arg)
    val_list[validx].message_ref = e_message_ref
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def set_page_field_validator_error_message_reference(project, pagenumber, pchange, widget_name, field_arg, validx, e_message_ref):
    "Set the error message reference on the validator at index validx within the validator list attached to the field, return the new page change"
    proj, page, val_list = _get_page_val_list(project, pagenumber, pchange, widget_name, field_arg)
    val_list[validx].message_ref = e_message_ref
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def set_section_field_validator_displaywidget(project, section_name, schange, widget_name, field_arg, validx, displaywidget):
    "Set the error message reference on the validator at index validx within the validator list attached to the field, return the new section change"
    proj, section, val_list = _get_section_val_list(project, section_name, schange, widget_name, field_arg)
    val_list[validx].displaywidget = displaywidget
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def set_page_field_validator_displaywidget(project, pagenumber, pchange, widget_name, field_arg, validx, displaywidget):
    "Set the displaywidget on the validator at index validx within the validator list attached to the field, return the new page change"
    proj, page, val_list = _get_page_val_list(project, pagenumber, pchange, widget_name, field_arg)
    val_list[validx].displaywidget = displaywidget
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def add_section_field_validator_allowed_value(project, section_name, schange, widget_name, field_arg, validx, allowed_value):
    "Add allowed value on the validator at index validx within the validator list attached to the field, return the new section change"
    proj, section, val_list = _get_section_val_list(project, section_name, schange, widget_name, field_arg)
    validator = val_list[validx]
    if not allowed_value:
        raise ServerError("A none empty string is required")
    if allowed_value in validator.allowed_values:
        raise ServerError("Allowed value already exists")
    lowval = allowed_value.lower()
    if lowval == "</script>":
        raise ServerError("Invalid allowed value : strings of this form may confuse javascript")
    if "\"" in lowval:
        raise ServerError("Invalid allowed value : strings of this form may confuse javascript")
    validator.allowed_values.append(allowed_value)
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def add_page_field_validator_allowed_value(project, pagenumber, pchange, widget_name, field_arg, validx, allowed_value):
    "Add allowed value on the validator at index validx within the validator list attached to the field, return the new page change"
    proj, page, val_list = _get_page_val_list(project, pagenumber, pchange, widget_name, field_arg)
    validator = val_list[validx]
    if not allowed_value:
        raise ServerError("A none empty string is required")
    if allowed_value in validator.allowed_values:
        raise ServerError("Allowed value already exists")
    lowval = allowed_value.lower()
    if lowval == "</script>":
        raise ServerError("Invalid allowed value : strings of this form may confuse javascript")
    if "\"" in lowval:
        raise ServerError("Invalid allowed value : strings of this form may confuse javascript")
    validator.allowed_values.append(allowed_value)
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def remove_section_field_validator_allowed_value(project, section_name, schange, widget_name, field_arg, validx, allowed_value_index):
    "Remove allowed value (given its index) on the validator at index validx within the validator list attached to the field, return the new section change"
    proj, section, val_list = _get_section_val_list(project, section_name, schange, widget_name, field_arg)
    value_list = val_list[validx].allowed_values
    if not value_list:
        raise ServerError("Allowed value list is empty")
    if (allowed_value_index >= 0) and (allowed_value_index < len(value_list)):
        del value_list[allowed_value_index]
    else:
        raise ServerError("Invalid value to remove")
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def remove_page_field_validator_allowed_value(project, pagenumber, pchange, widget_name, field_arg, validx, allowed_value_index):
    "Remove allowed value (given its index) on the validator at index validx within the validator list attached to the field, return the new page change"
    proj, page, val_list = _get_page_val_list(project, pagenumber, pchange, widget_name, field_arg)
    value_list = val_list[validx].allowed_values
    if not value_list:
        raise ServerError("Allowed value list is empty")
    if (allowed_value_index >= 0) and (allowed_value_index < len(value_list)):
        del value_list[allowed_value_index]
    else:
        raise ServerError("Invalid value to remove")
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def set_section_field_validator_argument(project, section_name, schange, widget_name, field_arg, validx, arg_name, arg_value):
    "Set an argument value on the validator at index validx within the validator list attached to the field, return the new section change"
    proj, section, val_list = _get_section_val_list(project, section_name, schange, widget_name, field_arg)
    validator = val_list[validx]
    try:
        validator[arg_name] = arg_value
    except Exception:
        raise ServerError("Invalid argument")
    # save the altered section, and return the section.change uuid
    return proj.add_section(section_name, section)


def set_page_field_validator_argument(project, pagenumber, pchange, widget_name, field_arg, validx, arg_name, arg_value):
    "Set an argument value (given its index) on the validator at index validx within the validator list attached to the field, return the new page change"
    proj, page, val_list = _get_page_val_list(project, pagenumber, pchange, widget_name, field_arg)
    validator = val_list[validx]
    try:
        validator[arg_name] = arg_value
    except Exception:
        raise ServerError("Invalid argument")
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)










"""Contains commonly used error display widgets"""

from string import Template

from .. import skiboot
from .. import tag
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class ErrorCode(Widget):
    """A div containing a paragraph, with an error code, and a second paragraph containing text,
       if an error is raised, then the text is changed
       for the error_message and the error_class given to the paragraph"""

    error_location = (1,0)

    arg_descriptions = {'para_text':FieldArg("text", "", jsonset=True),
                        'para_class':FieldArg("cssclass", ""),
                        'code_class':FieldArg("cssclass", ""),
                        'error_class':FieldArg("cssclass", ""),
                        'pre_line':FieldArg("boolean", True),
                        'code':FieldArg("integer", "0")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        para_text: The text appearing in the paragraph
        para_class: The class of the paragraph
        error_class: The class of the error text - which provides the appearance via CSS
                     replaces para_class on error.
        pre_line: If True, sets style="white-space: pre-line;" into the paragraph which preserves
                  new line breaks
        code: integer code number
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name='p')
        self[0][0] = ''
        self[1] = tag.Part(tag_name='p')
        # do not insert <br /> - leave that to pre_line
        self[1].linebreaks=False
        self[1][0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value('code_class'):
            self[0].update_attribs({"class":self.get_field_value('code_class')})
        if self.get_field_value("code"):
            self[0][0] = "Error code : " + str(self.get_field_value("code"))
        else:
            self[0][0] = "Error code : 0"
        if self.get_field_value("pre_line"):
            self[1].attribs={"style":"white-space: pre-line;"}
        if self.get_field_value('para_class'):
            self[1].update_attribs({"class":self.get_field_value('para_class')})
        # self[1][0] could be set by an error message
        if not self.error_status:
            self[1][0] = self.get_field_value("para_text")
        if self.error_status and self.get_field_value('error_class'):
            self[1].update_attribs({"class":self.get_field_value('error_class')})

            
    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets fieldvalues"""
        return self._make_fieldvalues('para_class', 'error_class')


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
    <p> <!-- with class code_class -->
        Error code : n
        <!-- with n set to the number given in field 'code' -->
    </p>
    <p style = "white-space: pre-line;">
    <!-- with class para_class, and style set if pre_line is True -->
    <!-- and class changed to error_class on error -->
        <!-- set with para_text, replaced by error message on error -->
    </p>
</div>"""




class ErrorDiv(Widget):
    """A div containing a div with paragraph containing error text - normally both div's are hidden.
       though the outer one can be displayed if required
       Can contain further html / widgets after the error div
       On displaying an error, the both div's are shown, and paragraph text set to error message."""

    _container = ((1,),)

    error_location = (0,0,0)

    arg_descriptions = {
                        'hide':FieldArg("boolean", True, jsonset=True),
                        'error_class':FieldArg("cssclass", ""),
                        'container_class':FieldArg("cssclass", '')
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        hide: If True, sets display: none; on the widget, can be set/unset via JSON file
              If False, or error sets display:block
        error_class: The CSS class of the paragraph
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # The location 1, is available as a container
        self[1] = tag.Part(tag_name="div")
        self[1][0] = ''

    def _build(self, page, ident_list, environ, call_data, lang):
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))
        if self.get_field_value('error_class'):
            self[0].attribs = {'class':self.get_field_value('error_class')}
        if self.error_status:
            self.update_attribs({"style":"display: block;"})
            self[0].del_one_attrib("style")
        # the div holding the container
        if self.get_field_value('container_class'):
            self[1].attribs = {"class": self.get_field_value('container_class')}

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class, normally hidden -->
  <div>  !-- class set to error_class -->
    <p> <!-- error message appears in this paragraph --> </p>
  </div>
  <div>  <!-- this div has the class attribute set to container_class -->
    <!-- container 0 for further html -->
  </div>
</div>"""


class ErrorPara(Widget):
    """A paragraph containing error text - normally hidden, displayed on error"""

    error_location = 0

    arg_descriptions = {
                        'hide':FieldArg("boolean", True, jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        hide: If True, sets display: none; on the widget, can be set/unset via JSON file
              If False, or error sets display:block
        """
        Widget.__init__(self, name=name, tag_name="p", brief=brief, **field_args)
        self[0] = self.error_message

    def _build(self, page, ident_list, environ, call_data, lang):
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<p>  <!-- with widget id and class widget_class, normally hidden,
          shown when an error is displayed. -->
  <!-- error message appears in this paragraph -->
</p>
"""


class ErrorClear1(Widget):
    """A div - normally used for modal background containing a div with a paragraph
         of text or error message and a clear button with three optional get fields.
         The 'clear' button hides the widget without making a call if javascript is available
         if not, then the link is called."""

    error_location = (0,0,0,0)

    arg_descriptions = {'hide':FieldArg("boolean", True, jsonset=True),
                        'para_text':FieldArg("text", "", jsonset=True),
                        'pre_line':FieldArg("boolean", True),
                        'boxdiv_class':FieldArg("cssclass", ""),
                        'error_class':FieldArg("cssclass", ""),
                        'buttondiv_class':FieldArg("cssclass", ""),
                        'buttondiv_style':FieldArg("cssstyle", ""),
                        'button_class':FieldArg("cssclass", ""),
                        'link_ident':FieldArg("url", ''),
                        'get_field1':FieldArg("text", "", valdt=True, jsonset=True),
                        'get_field2':FieldArg("text","", valdt=True, jsonset=True),
                        'get_field3':FieldArg("text","", valdt=True, jsonset=True),
                        'button_text':FieldArg("text", "Clear"),
            }


    def __init__(self, name=None, brief='', **field_args):
        """
        hide: If True, sets display: none; on the widget, can be set/unset via JSON file
              If False, or error sets display:block
        para_text: The text appearing in the paragraph, replaced on error by error message
        pre_line: If True, sets style="white-space: pre-line;" into the paragraph which preserves
                  new line breaks
        boxdiv_class: class of the box holding paragraph and button
        error_class: The CSS class of the div holding the paragraph
        buttondiv_class: The class of the div holding the button
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="div")
        # The location 0,0 is the div holding the text paragraph
        self[0][0] = tag.Part(tag_name="div")
        self[0][0][0] = tag.Part(tag_name="p")
        # do not set any linebreaks, leave that to pre-line
        self[0][0][0].linebreaks = False
        self[0][0][0][0] = ''
        # div holding buttons
        self[0][1] = tag.Part(tag_name="div")
        self[0][1][0] = tag.Part(tag_name="a", attribs={"role":"button"})

    def _build(self, page, ident_list, environ, call_data, lang):
        "build the box"
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))
        if self.get_field_value("boxdiv_class"):
            self[0].update_attribs({"class":self.get_field_value('boxdiv_class')})
        if self.get_field_value("error_class"):
            self[0][0].update_attribs({"class":self.get_field_value('error_class')})
        if self.get_field_value("buttondiv_class"):
            self[0][1].update_attribs({"class":self.get_field_value('buttondiv_class')})
        if self.get_field_value("buttondiv_style"):
            self[0][1].update_attribs({'style':self.get_field_value("buttondiv_style")})

        if self.get_field_value("pre_line"):
            self[0][0][0].attribs={"style":"white-space: pre-line;"}

        if not self.error_status:
            self[0][0][0][0] = self.get_field_value("para_text")

        # button
        if self.get_field_value('button_class'):
            self[0][1][0].update_attribs({"class":self.get_field_value('button_class')})
        if not self.get_field_value("link_ident"):
            self[0][1][0][0] = "Warning: broken link"
        else:
            url = skiboot.get_url(self.get_field_value("link_ident"), proj_ident=page.proj_ident)
            if not url:
                self[0][1][0][0] = "Warning: broken link"
            else:
                if self.get_field_value("button_text"):
                    self[0][1][0][0] = self.get_field_value("button_text")
                else:
                    self[0][1][0][0] = url
                # create a url for the href
                get_fields = {self.get_formname("get_field1"):self.get_field_value("get_field1"),
                                            self.get_formname("get_field2"):self.get_field_value("get_field2"),
                                            self.get_formname("get_field3"):self.get_field_value("get_field3")}
                url = self.make_get_url(page, url, get_fields, True)
                self[0][1][0].update_attribs({"href": url})


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a click event handler on the a button"""
        jscript = """  $("#{ident} a").click(function (e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident = self.get_id())
        return jscript


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div> <!-- With boxdiv_class -->
    <div> <!-- With class set by error_class -->
      <p style = "white-space: pre-line;"> <!-- style set if pre_line is True -->
        <!-- para_text or error message appears in this paragraph -->
      </p>
    </div>
    <div>    <!-- with class set by buttondiv_class and style by buttondiv_style -->
      <a role="button" href="#">
        <!-- With class set by button_class, and the href link will be the url of the link_ident with the three get_fields -->
        <!-- the button will show the button_text -->
      </a>
    </div>
  </div>
</div>"""


class ErrorClear2(Widget):
    """A div - normally used for modal background containing a div with a paragraph
         of error message and an X button with three optional get fields.
         The 'X' button hides the widget without making a call if javascript is available
         if not, then the link is called."""

    error_location = (0,1,0,0)

    arg_descriptions = {'hide':FieldArg("boolean", True, jsonset=True),
                        'pre_line':FieldArg("boolean", True),
                        'boxdiv_class':FieldArg("cssclass", ""),
                        'error_class':FieldArg("cssclass", ""),
                        'buttondiv_class':FieldArg("cssclass", ""),
                        'buttondiv_style':FieldArg("cssstyle", ""),
                        'button_class':FieldArg("cssclass", ""),
                        'link_ident':FieldArg("url", ''),
                        'get_field1':FieldArg("text", "", valdt=True, jsonset=True),
                        'get_field2':FieldArg("text","", valdt=True, jsonset=True),
                        'get_field3':FieldArg("text","", valdt=True, jsonset=True)
            }


    def __init__(self, name=None, brief='', **field_args):
        """
        hide: If True, sets display: none; on the widget, can be set/unset via JSON file
              If False, or error sets display:block
        pre_line: If True, sets style="white-space: pre-line;" into the paragraph which preserves
                  new line breaks
        boxdiv_class: class of the box holding paragraph and button
        error_class: The CSS class of the div holding the paragraph
        buttondiv_class: The class of the div holding the button
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="div")
        # div holding X button
        self[0][0] = tag.Part(tag_name="div")
        self[0][0][0] = tag.Part(tag_name="a", attribs={"role":"button"})
        self[0][0][0][0] = tag.HTMLSymbol("&times;")
        # The location 0,1 is the div holding the text
        self[0][1] = tag.Part(tag_name="div")
        self[0][1][0] = tag.Part(tag_name="p")
        # do not set any linebreaks, leave that to pre-line
        self[0][1][0].linebreaks = False
        self[0][1][0][0] = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the box"
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))
        if self.get_field_value("boxdiv_class"):
            self[0].update_attribs({"class":self.get_field_value('boxdiv_class')})
        # buttondiv
        if self.get_field_value("buttondiv_class"):
            self[0][0].update_attribs({"class":self.get_field_value('buttondiv_class')})
        if self.get_field_value("buttondiv_style"):
            self[0][0].update_attribs({'style':self.get_field_value("buttondiv_style")})
        # inner div
        if self.get_field_value("error_class"):
            self[0][1].update_attribs({"class":self.get_field_value('error_class')})
        # paragraph
        if self.get_field_value("pre_line"):
            self[0][1][0].attribs={"style":"white-space: pre-line;"}
        if not self.error_status:
            self[0][1][0][0] = self.get_field_value("show_error")
        # insert an id into the paragraph for setting the error text
        self[0][1][0].insert_id()
        # button
        if self.get_field_value('button_class'):
            self[0][0][0].update_attribs({"class":self.get_field_value('button_class')})
        if not self.get_field_value("link_ident"):
            self[0][1][0][0] = "Warning: broken link"
        else:
            url = skiboot.get_url(self.get_field_value("link_ident"), proj_ident=page.proj_ident)
            if url:
                # create a url for the href
                get_fields = {self.get_formname("get_field1"):self.get_field_value("get_field1"),
                              self.get_formname("get_field2"):self.get_field_value("get_field2"),
                              self.get_formname("get_field3"):self.get_field_value("get_field3")}
                url = self.make_get_url(page, url, get_fields, True)
                self[0][0][0].update_attribs({"href": url})
            else:
                self[0][1][0][0] = "Warning: broken link"


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a click event handler on the a button"""
        jscript = """  $("#{ident} a").click(function (e) {{
    SKIPOLE.widgets['{ident}'].eventfunc(e);
    }});
""".format(ident = self.get_id())
        # return this javascript and the paragraph id
        return jscript + self._make_fieldvalues(para_id=self[0][1][0].get_id())


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div> <!-- With boxdiv_class -->
    <div>    <!-- with class set by buttondiv_class and style by buttondiv_style -->
      <a role="button" href="#">
        <!-- With class set by button_class, and the href link will be the url of the link_ident with the three get_fields -->
        <!-- the button will show the &times; symbol -->
      </a>
    </div>
    <div> <!-- With class set by error_class -->
      <p style = "white-space: pre-line;"> <!-- style set if pre_line is True -->
        <!-- error message appears in this paragraph -->
      </p>
    </div>
  </div>
</div>"""






import os, re

from skipole import ValidateError, FailPage, ServerError

from skipole import skilift
from skipole.skilift import editfolder, fromjson



# a search for anything none-alphanumeric, not a dot and not a underscore and not an hyphen
_ANDH = re.compile('[^\w\.\-]')



def _get_folder_info(project, folder):
    "Given a folder ident string such as number or 'project,number' or 'project_number' return FolderInfo, folder_url"
    try:
        foldernumber = skilift.get_itemnumber(project, folder)
        if foldernumber is None:
            raise FailPage(message="Parent folder not recognised")
        folder_info = skilift.folder_info(project, foldernumber)
        folder_url = skilift.page_path(project, foldernumber)
    except ServerError as e:
        raise FailPage(message=e.message)
    return folder_info, folder_url


def retrieve_add_folder(skicall):
    "Fill in the add a folder page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    # parent is the folder a new folder is to be added to
    # the value in call_data is the string ident submitted by the ftree add_folder button
    # or by a value in session_data
    if 'parent' in call_data:
        parent_info, parent_url = _get_folder_info(project, call_data['parent'])
    elif 'add_to_foldernumber' in call_data:
        parent_info, parent_url = _get_folder_info(project, call_data['add_to_foldernumber'])
    else:
        raise FailPage(message = "Parent folder missing")

    page_data[("adminhead","page_head","large_text")] = "Add folder to : %s" % (parent_url,)

    page_data[('staticpath','input_text')] = os.path.join(project, 'static')

    page_data[('newfolderform','parent')] = project+","+str(parent_info.number)
    call_data['add_to_foldernumber'] = parent_info.number

    # st1: new folder name
    if 'new_folder' in call_data:
        page_data[('foldername','new_folder')] = call_data['new_folder']

    # cb1: restricted checkbox
    if ('checkbox' in call_data) and call_data['checkbox']:
        page_data[('cb1','checked')] = True
    else:
        page_data[('cb1','checked')] = False
    if parent_info.restricted:
        page_data[('cb1','show_restricted')] = False
    else:
        page_data[('cb1','show_restricted')] = True

    # it1: text input for folder brief
    if ('folder_brief' in call_data) and call_data['folder_brief']:
        page_data[('it1','folder_brief')] = call_data['folder_brief']

    # it2: folder ident number
    if 'folder_ident_number' in call_data:
        page_data[('it2','folder_ident_number')] = str(call_data['folder_ident_number'])
    else:
        page_data[('it2','folder_ident_number')] = str(skilift.next_ident_number(project))


def submit_addfolder(skicall):
    """ Creates a folder by making a dictionary similar to:

    {
     "name":"folder_name",
     "ident":999,
     "brief":"brief description of the folder",
     "restricted":False
    }

    And then calling editfolder.make_new_folder.

    Also calls _make_static_folder if folderpath is in call data

"""

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    folder_dict = {}

    if 'parent' not in call_data:
        raise FailPage(message = "Parent folder missing")
    # the parent value in call_data is the string ident submitted by the button
    parentinfo, parent_url = _get_folder_info(project, call_data['parent'])

    # parentinfo is a named tuple with members
    # 'name', 'number', 'restricted', 'brief', 'contains_pages', 'contains_folders'

    if ('new_folder' not in call_data) or ('checkbox' not in call_data) or ('folder_brief' not in call_data) or ('folder_ident_number' not in call_data):
        raise FailPage("New folder information missing")

    try:
        folder_ident_number = int(call_data['folder_ident_number'])
    except Exception:
        raise FailPage("The Folder Ident number must be an integer")
    if folder_ident_number<1:
        raise FailPage("The Folder Ident number must be a positive integer greater than zero")
    folder_dict["ident"] = folder_ident_number
    folder_dict["brief"] = call_data['folder_brief']
    folder_dict["restricted"] = call_data['checkbox']

    new_folder_name = call_data['new_folder']
    # check name is alphanumric or underscore or dot or hyphen only
    if _ANDH.search(new_folder_name):
        raise FailPage(message = "Folder names must be alphanumric and may also have dots or underscores or hyphens")
    folder_dict["name"] = new_folder_name


    if 'folderpath' in call_data and call_data['folderpath']:
        folderpath = call_data['folderpath'].strip()
        folderpath = folderpath.strip('/')
        folderpath = folderpath.strip('\\')
        if not folderpath:
            raise FailPage("Sorry, the given static folder is invalid.")
        fullpath = os.path.join(skilift.get_projectfiles_dir(project), folderpath)
        if not os.path.isdir(fullpath):
            raise FailPage("Sorry, the given static folder location cannot be found.")
        if not call_data['folder_brief']:
            folder_dict["brief"] = "Link to %s" % folderpath
    else:
        folderpath = None
        fullpath = None

    # folderpath is the server folder path relative to projectfiles
    # fullpath is the absolute server folder path

    try:
        # create the folder
        editfolder.make_new_folder(project, parentinfo.number, folder_dict)
    except ServerError as e:
        raise FailPage(message = e.message)

    if fullpath:
        # add subfolders and file pages
        _make_static_folder(project, folder_dict, fullpath, folderpath)
        call_data['status'] = 'Static folder tree added'
        return

    call_data['status'] = 'New folder %s added.' % (parent_url + folder_dict["name"] + '/',)


def _make_static_folder(project, folder_dict, fullpath, folderpath):
    """Creates containing sub folders and Filepages pointing to static server files

    folderpath is the server folder path relative to projectfiles
    fullpath is the absolute server folder path
"""
    try:
        # loads everything under folderpath as Folders and FilePages
        # ident_dict maps folderpath to newly created folder ident numbers
        ident_dict = {}
        ident_dict[folderpath] = folder_dict["ident"]
        ident = folder_dict["ident"]
        ident_number_list = skilift.ident_numbers(project)
        for root, dirs, files in os.walk(fullpath):
            #fpath = root[len(skilift.get_projectdir(project))+1:]
            fpath = root[len(skilift.get_projectfiles_dir(project))+1:]
            parent_ident = ident_dict[fpath]
            if files:
                # create files
                for filename in files:
                    new_filepath=os.path.join(fpath, filename)
                    new_page_dict = {"name":filename,
                                     "brief":"Link to %s" % (new_filepath,),
                                     "FilePage": {
                                         "filepath": new_filepath,
                                         }
                                     }
                    if ident:
                        ident +=1
                        if ident not in ident_number_list:
                            new_page_dict["ident"] = ident
                    editfolder.make_new_page(project, parent_ident, new_page_dict)
            if dirs:
                # create folders
                for foldername in dirs:
                    new_folderpath=os.path.join(fpath, foldername)
                    new_folder_dict = {"name":foldername,
                                       "brief":"Link to %s" % (new_folderpath,),
                                       "restricted":False
                                      }
                    if ident:
                        ident +=1
                        if ident not in ident_number_list:
                            new_folder_dict["ident"] = ident
                    ident_dict[new_folderpath] = editfolder.make_new_folder(project, parent_ident, new_folder_dict)
    except ServerError as e:
        raise FailPage(e.message)



def submit_upload_folder(skicall):
    "Copy a folder from uploaded file"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    # add_to_foldernumber is the folder a new folder is to be added to
    if 'add_to_foldernumber' not in call_data:
        raise FailPage(message = "Parent folder missing")

    # get submitted data for new folder
    try:
        addident = int(call_data['addident'])
    except Exception:
        raise FailPage(message = "Addition integer is invalid")
    importname = call_data['importname']
    uploadfile = call_data['uploadfile']
    json_string = uploadfile.decode(encoding='utf-8')
    # create the folder
    try:
        # note: restricted is set to False
        fromjson.create_folder(project, call_data['add_to_foldernumber'], addident, importname, False, json_string)
    except ServerError as e:
        raise FailPage(message = e.message, widget='import_folder')
    del call_data['add_to_foldernumber']
    call_data['status'] = 'New folder and contents added'







"""Contains widgets for uploading files"""


from .. import skiboot, tag, excepts
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class SubmitUploadFile1(Widget):
    """Defines a form with a file input field"""

    error_location = (0,0,0)

    arg_descriptions = {'filebutton_label':FieldArg("text", ''),
                        'filebuttonlabel_class':FieldArg("cssclass", ''),
                        'filebutton_class':FieldArg("cssclass", ''),
                        'action':FieldArg("url", '', valdt=True),
                        'submitbutton_label':FieldArg("text", ''),
                        'submitbuttonlabel_class':FieldArg("cssclass", ''),
                        'submitbutton_text':FieldArg("text",'Submit'),
                        'submitbutton_class':FieldArg("cssclass", ''),
                        'fileinputdiv_class':FieldArg("cssclass", ''),
                        'submitinputdiv_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", ''),
                        'required':FieldArg("boolean", False),
                        'hide':FieldArg("boolean", False, jsonset=True),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        filebutton_label: The text displayed to the left of the file button
        filebuttonlabel_class: The css class of the label
        filebutton_class: The css class of the file browse button
        action: The ident or label to send this file to, name of this field is used as widgfield fieldname
        submitbutton_text: The text on the button
        submitbutton_label: The text displayed to the left of the submit button
        submitbuttonlabel_class: The css class of the label
        submitbutton_class: The css class of the submit button
        fileinputdiv_class: the class attribute of the div which contains the file label and button
        submitinputdiv_class: the class attribute of the div which contains the submit label and button
        error_class: The class applied to the paragraph containing the error message on error.
        required: Set True to put the 'required' flag into the input field
        hide: If True, widget is hidden
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # hidden error
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post", "enctype":"multipart/form-data"})
        # div containing file label and button
        self[1][0] = tag.Part(tag_name='div')
        # the label
        self[1][0][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the file button field
        self[1][0][1] = tag.ClosedPart(tag_name="input", attribs ={"type":"file"})
        # div containing submit label and button
        self[1][1] = tag.Part(tag_name='div')
        # the label
        self[1][1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the submit button
        self[1][1][1] = tag.ClosedPart(tag_name="input", attribs ={"type":"submit"})


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if not self.get_field_value("action"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = skiboot.get_url(self.get_field_value("action"),  proj_ident=page.proj_ident)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[1].update_attribs({"action": actionurl})

        # file button
        # the div holding file label and button
        if self.get_field_value('fileinputdiv_class'):
            self[1][0].attribs = {"class": self.get_field_value('fileinputdiv_class')}
        if self.get_field_value('filebutton_label'):
            self[1][0][0][0] = self.get_field_value('filebutton_label')
        if self.get_field_value('filebuttonlabel_class'):
            self[1][0][0].attribs = {"class": self.get_field_value('filebuttonlabel_class')}

        # set an id in the file button for the 'label for' tag
        self[1][0][1].insert_id()

        self[1][0][1].update_attribs({"name":self.get_formname('action')})

        if self.get_field_value('required'):
            self[1][0][1].update_attribs({"required":"required"})
        if self.get_field_value('filebutton_class'):
            self[1][0][1].update_attribs({"class": self.get_field_value('filebutton_class')})


        # set the label 'for' attribute
        self[1][0][0].update_attribs({'for':self[1][0][1].get_id()})

        # submit button
        # the div holding submit label and button
        if self.get_field_value('submitinputdiv_class'):
            self[1][1].attribs = {"class": self.get_field_value('submitinputdiv_class')}
        if self.get_field_value('submitbutton_label'):
            self[1][1][0][0] = self.get_field_value('submitbutton_label')
        if self.get_field_value('submitbuttonlabel_class'):
            self[1][1][0].attribs = {"class": self.get_field_value('submitbuttonlabel_class')}

        # submit button
        self[1][1][1].update_attribs({"value":self.get_field_value('submitbutton_text')})
        # set an id in the submit button for the 'label for' tag
        self[1][1][1].insert_id()

        if self.get_field_value('submitbutton_class'):
            self[1][1][1].update_attribs({"class": self.get_field_value('submitbutton_class')})

        # set the label 'for' attribute
        self[1][1][0].update_attribs({'for':self[1][1][1].get_id()})

        # add ident and four hidden fields
        self.add_hiddens(self[1], page)



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div>  <!-- div hidden when no error is displayed, with class set to error_class on error -->
    <p> <!-- error message appears in this paragraph --> </p>
  </div>
  <form role="form" method="post" enctype="multipart/form-data"> <!-- action attribute set to action field -->
    <div> <!-- class attribute set to fileinputdiv_class -->
      <label> <!-- with class set to filebuttonlabel_class and content to filebutton_label -->
      </label>
      <input type="file" />  <!-- class attribute set to filebutton_class -->
    </div>
    <div> <!-- class attribute set to submitinputdiv_class -->
      <label> <!-- with class set to submitbuttonlabel_class and content to submitbutton_label -->
      </label>
      <input type="submit" /> <!-- button value set to submitbutton_text and class to submitbutton_class-->
    </div>
    <!-- hidden fields -->
  </form>
</div>"""


class UploadFile1(Widget):
    """Defines a file input field"""

    error_location = (0,0,0)

    arg_descriptions = {'label':FieldArg("text", 'Pick a file to upload:'),
                        'inputname':FieldArg("text", '', valdt=True),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'filebutton_class':FieldArg("cssclass", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", '')
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        label: The text displayed to the left of the file button
        inputname: the field name of the widgfield which is submitted
        label_class: The css class of the label
        filebutton_class: The css class of the file browse button
        inputdiv_class: The css class of the div containing the label and input field
        error_class: The class applied to the paragraph containing the error message on error.
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # inputdiv
        self[1] = tag.Part(tag_name="div")
        # the label
        self[1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the file button field
        self[1][1] = tag.ClosedPart(tag_name="input", attribs ={"type":"file"})


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")

        # file button
        # the div holding file label and button
        if self.get_field_value('inputdiv_class'):
            self[1].attribs = {"class": self.get_field_value('inputdiv_class')}
        if self.get_field_value('label'):
            self[1][0][0] = self.get_field_value('label')
        if self.get_field_value('label_class'):
            self[1][0].attribs = {"class": self.get_field_value('label_class')}
        if self.get_field_value('label_style'):
            self[1][0].attribs = {"style": self.get_field_value('label_style')}

        # set an id in the file button for the 'label for' tag
        self[1][1].insert_id()
        self[1][1].update_attribs({"name":self.get_formname('inputname')})
        if self.get_field_value('filebutton_class'):
            self[1][1].update_attribs({"class": self.get_field_value('filebutton_class')})
        # set the label 'for' attribute
        self[1][0].update_attribs({'for':self[1][1].get_id()})


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <div>  <!-- class attribute set to inputdiv_class -->
    <label> <!-- with class set to label_class -->
            <!-- content set to label -->
    </label>
    <input type="file" />  <!-- class attribute set to filebutton_class -->
  </div>
</div>"""


